# coding=utf-8

# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Default operators defined on any (reasonable) space."""

from __future__ import absolute_import, division, print_function

from copy import copy

from numbers import Number
import numpy as np

from odl.core.operator.operator import Operator
from odl.core.set import ComplexNumbers, Field, LinearSpace, RealNumbers
from odl.core.set.space import LinearSpaceElement
from odl.core.space import ProductSpace
from odl.core.array_API_support import sqrt, conj

__all__ = ('ScalingOperator', 'ZeroOperator', 'IdentityOperator',
           'LinCombOperator', 'MultiplyOperator', 'PowerOperator',
           'InnerProductOperator', 'NormOperator', 'DistOperator',
           'ConstantOperator', 'RealPart', 'ImagPart', 'ComplexEmbedding',
           'ComplexModulus', 'ComplexModulusSquared')


class ScalingOperator(Operator):

    """Operator of multiplication with a scalar.

    Implements::

        ScalingOperator(s)(x) == s * x
    """

    def __init__(self, domain, scalar):
        """Initialize a new instance.

        Parameters
        ----------
        domain : `LinearSpace` or `Field`
            Set of elements on which this operator acts.
        scalar : ``domain.field`` element
            Fixed scaling factor of this operator.

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> vec = r3.element([1, 2, 3])
        >>> out = r3.element()
        >>> op = ScalingOperator(r3, 2.0)
        >>> op(vec, out)  # In-place, Returns out
        rn(3).element([ 2.,  4.,  6.])
        >>> out
        rn(3).element([ 2.,  4.,  6.])
        >>> op(vec)  # Out-of-place
        rn(3).element([ 2.,  4.,  6.])
        """
        if not isinstance(domain, (LinearSpace, Field)):
            raise TypeError('`domain` {!r} not a `LinearSpace` or `Field` '
                            'instance'.format(domain))

        super(ScalingOperator, self).__init__(domain, domain, linear=True)
        self.__scalar = domain.field.element(scalar)

    @property
    def scalar(self):
        """Fixed scaling factor of this operator."""
        return self.__scalar

    def _call(self, x, out=None):
        """Scale ``x`` and write to ``out`` if given."""
        if out is None:
            out = self.scalar * x
        else:
            out.lincomb(self.scalar, x)
        return out

    @property
    def inverse(self):
        """Return the inverse operator.

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> vec = r3.element([1, 2, 3])
        >>> op = ScalingOperator(r3, 2.0)
        >>> inv = op.inverse
        >>> inv(op(vec)) == vec
        True
        >>> op(inv(vec)) == vec
        True
        """
        if self.scalar == 0.0:
            raise ZeroDivisionError('scaling operator not invertible for '
                                    'scalar==0')
        return ScalingOperator(self.domain, 1.0 / self.scalar)

    @property
    def adjoint(self):
        """Adjoint, given as scaling with the conjugate of the scalar.

        Examples
        --------
        In the real case, the adjoint is the same as the operator:

        >>> r3 = odl.rn(3)
        >>> x = r3.element([1, 2, 3])
        >>> op = ScalingOperator(r3, 2)
        >>> op(x)
        rn(3).element([ 2.,  4.,  6.])
        >>> op.adjoint(x)  # The same
        rn(3).element([ 2.,  4.,  6.])

        In the complex case, the scalar is conjugated:

        >>> c3 = odl.cn(3)
        >>> x_complex = c3.element([1, 1j, 1-1j])
        >>> op = ScalingOperator(c3, 1+1j)
        >>> expected_op = ScalingOperator(c3, 1-1j)
        >>> op.adjoint(x_complex)
        cn(3).element([ 1.-1.j,  1.+1.j,  0.-2.j])
        >>> expected_op(x_complex)  # The same
        cn(3).element([ 1.-1.j,  1.+1.j,  0.-2.j])

        Returns
        -------
        adjoint : `ScalingOperator`
            ``self`` if `scalar` is real, else `scalar` is conjugated.
        """
        if complex(self.scalar).imag == 0.0:
            return self
        else:
            return ScalingOperator(self.domain, self.scalar.conjugate())

    def norm(self, estimate=False, **kwargs):
        """Return the operator norm of this operator.

        Parameters
        ----------
        estimate, kwargs : bool
            Ignored. Present to conform with base-class interface.

        Returns
        -------
        norm : float
            The operator norm, absolute value of `scalar`.

        Examples
        --------
        >>> spc = odl.rn(3)
        >>> scaling = odl.ScalingOperator(spc, 3.0)
        >>> scaling.norm(True)
        3.0
        """
        return np.abs(self.scalar)

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r}, {!r})'.format(self.__class__.__name__,
                                       self.domain, self.scalar)

    def __str__(self):
        """Return ``str(self)``."""
        return '{} * I'.format(self.scalar)


class IdentityOperator(ScalingOperator):

    """Operator mapping each element to itself.

    Implements::

        IdentityOperator()(x) == x
    """

    def __init__(self, space):
        """Initialize a new instance.

        Parameters
        ----------
        space : `LinearSpace`
            Space of elements which the operator is acting on.
        """
        super(IdentityOperator, self).__init__(space, 1)

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r})'.format(self.__class__.__name__, self.domain)

    def __str__(self):
        """Return ``str(self)``."""
        return "I"


class LinCombOperator(Operator):

    """Operator mapping two space elements to a linear combination.

    Implements::

        LinCombOperator(a, b)([x, y]) == a * x + b * y
    """

    def __init__(self, space, a, b):
        """Initialize a new instance.

        Parameters
        ----------
        space : `LinearSpace`
            Space of elements which the operator is acting on.
        a, b : ``space.field`` elements
            Scalars to multiply ``x[0]`` and ``x[1]`` with, respectively.

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> r3xr3 = odl.ProductSpace(r3, r3)
        >>> xy = r3xr3.element([[1, 2, 3], [1, 2, 3]])
        >>> z = r3.element()
        >>> op = LinCombOperator(r3, 1.0, 1.0)
        >>> op(xy, out=z)  # Returns z
        rn(3).element([ 2.,  4.,  6.])
        >>> z
        rn(3).element([ 2.,  4.,  6.])
        """
        domain = ProductSpace(space, space)
        super(LinCombOperator, self).__init__(domain, space, linear=True)
        self.a = a
        self.b = b

    def _call(self, x, out=None):
        """Linearly combine ``x`` and write to ``out`` if given."""
        if out is None:
            out = self.range.element()
        out.lincomb(self.a, x[0], self.b, x[1])
        return out

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r}, {!r}, {!r})'.format(self.__class__.__name__,
                                             self.range, self.a, self.b)

    def __str__(self):
        """Return ``str(self)``."""
        return "{}*x + {}*y".format(self.a, self.b)


class MultiplyOperator(Operator):

    """Operator multiplying by a fixed space or field element.

    Implements::

        MultiplyOperator(y)(x) == x * y

    Here, ``y`` is a `LinearSpaceElement` or `Field` element and
    ``x`` is a `LinearSpaceElement`.
    Hence, this operator can be defined either on a `LinearSpace` or on
    a `Field`. In the first case it is the pointwise multiplication,
    in the second the scalar multiplication.
    """

    def __init__(self, multiplicand, domain=None, range=None):
        """Initialize a new instance.

        Parameters
        ----------
        multiplicand : `LinearSpaceElement` or scalar
            Value to multiply by.
        domain : `LinearSpace` or `Field`, optional
            Set to which the operator can be applied.
            Default: ``multiplicand.space``.
        range : `LinearSpace` or `Field`, optional
            Set to which the operator maps. Default: ``multiplicand.space``.

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> x = r3.element([1, 2, 3])

        Multiply by vector:

        >>> op = MultiplyOperator(x)
        >>> op(x)
        rn(3).element([ 1.,  4.,  9.])
        >>> out = r3.element()
        >>> op(x, out)
        rn(3).element([ 1.,  4.,  9.])

        Multiply by scalar:

        >>> op2 = MultiplyOperator(x, domain=r3.field)
        >>> op2(3)
        rn(3).element([ 3.,  6.,  9.])
        >>> out = r3.element()
        >>> op2(3, out)
        rn(3).element([ 3.,  6.,  9.])
        """
        # TODO: handle the complex conversion case better.
        if not isinstance(multiplicand, LinearSpaceElement):
            assert domain is not None or range is not None
            if domain is None:
                domain = range
            if range is None:
                range = domain
            assert isinstance(multiplicand, Number)

        if domain is None:
            domain = multiplicand.space

        if range is None:
            range = multiplicand.space

        super(MultiplyOperator, self).__init__(domain, range, linear=True)

        self.__multiplicand = multiplicand
        self.__domain_is_field = isinstance(domain, Field)
        self.__range_is_field = isinstance(range, Field)

    @property
    def multiplicand(self):
        """Value to multiply by."""
        return self.__multiplicand

    def _call(self, x, out=None):
        """Multiply ``x`` and write to ``out`` if given."""
        if out is None:
            return x * self.multiplicand
        elif not self.__range_is_field:
            if self.__domain_is_field:
                out.lincomb(x, self.multiplicand)
            else:
                out.assign(self.multiplicand * x)
        else:
            raise ValueError('can only use `out` with `LinearSpace` range')

    @property
    def adjoint(self):
        """Adjoint of this operator.

        Returns
        -------
        adjoint : `InnerProductOperator` or `MultiplyOperator`
            If the domain of this operator is the scalar field of a
            `LinearSpace` the adjoint is the inner product with ``y``,
            else it is the multiplication with ``y``.

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> x = r3.element([1, 2, 3])

        Multiply by a space element:

        >>> op = MultiplyOperator(x)
        >>> out = r3.element()
        >>> op.adjoint(x)
        rn(3).element([ 1.,  4.,  9.])

        Multiply scalars with a fixed vector:

        >>> op2 = MultiplyOperator(x, domain=r3.field)
        >>> op2.adjoint(x)
        14.0

        Multiply vectors with a fixed scalar:

        >>> op2 = MultiplyOperator(3.0, domain=r3, range=r3)
        >>> op2.adjoint(x)
        rn(3).element([ 3.,  6.,  9.])

        Multiplication operator with complex space:

        >>> c3 = odl.cn(3)
        >>> x_complex = c3.element([1, 1j, 1-1j])
        >>> op3 = MultiplyOperator(x_complex)
        >>> op3.adjoint.multiplicand
        cn(3).element([ 1.-0.j,  0.-1.j,  1.+1.j])
        """
        if self.__domain_is_field:
            if isinstance(self.domain, RealNumbers):
                return InnerProductOperator(self.multiplicand)
            elif isinstance(self.domain, ComplexNumbers):
                return InnerProductOperator(self.multiplicand.conjugate())
            else:
                raise NotImplementedError(
                    'adjoint not implemented for domain{!r}'
                    ''.format(self.domain))
        elif self.domain.is_complex:
            return MultiplyOperator(conj(self.multiplicand),
                                    domain=self.range, range=self.domain)
        else:
            return MultiplyOperator(self.multiplicand,
                                    domain=self.range, range=self.domain)

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r})'.format(self.__class__.__name__, self.multiplicand)

    def __str__(self):
        """Return ``str(self)``."""
        return "x * {}".format(self.y)


class PowerOperator(Operator):

    """Operator taking a fixed power of a space or field element.

    Implements::

        PowerOperator(p)(x) == x ** p

    Here, ``x`` is a `LinearSpaceElement` or `Field` element and ``p`` is
    a number. Hence, this operator can be defined either on a
    `LinearSpace` or on a `Field`.
    """

    def __init__(self, domain, exponent):
        """Initialize a new instance.

        Parameters
        ----------
        domain : `LinearSpace` or `Field`
            Set of elements on which the operator can be applied.
        exponent : float
            Exponent parameter of the power function applied to an element.

        Examples
        --------
        Use with vectors

        >>> op = PowerOperator(odl.rn(3), exponent=2)
        >>> op([1, 2, 3])
        rn(3).element([ 1.,  4.,  9.])

        or scalars

        >>> op = PowerOperator(odl.RealNumbers(), exponent=2)
        >>> op(2.0)
        4.0
        """
        super(PowerOperator, self).__init__(
            domain, domain, linear=(exponent == 1))
        self.__exponent = float(exponent)
        self.__domain_is_field = isinstance(domain, Field)

    @property
    def exponent(self):
        """Power of the input element to take."""
        return self.__exponent

    def _call(self, x, out=None):
        """Take the power of ``x`` and write to ``out`` if given."""
        if out is None:
            return x ** self.exponent
        elif self.__domain_is_field:
            raise ValueError('cannot use `out` with field')
        else:
            out.assign(x)
            out **= self.exponent

    def derivative(self, point):
        """Derivative of this operator.

            ``PowerOperator(p).derivative(y)(x) == p * y ** (p - 1) * x``

        Parameters
        ----------
        point : `domain` element
            The point in which to take the derivative

        Returns
        -------
        derivative : `Operator`
            The derivative in ``point``

        Examples
        --------
        Use on vector spaces:

        >>> op = PowerOperator(odl.rn(3), exponent=2)
        >>> dop = op.derivative(op.domain.element([1, 2, 3]))
        >>> dop([1, 1, 1])
        rn(3).element([ 2.,  4.,  6.])

        Use with scalars:

        >>> op = PowerOperator(odl.RealNumbers(), exponent=2)
        >>> dop = op.derivative(2.0)
        >>> dop(2.0)
        8.0
        """
        return self.exponent * MultiplyOperator(point ** (self.exponent - 1),
                                                domain=self.domain,
                                                range=self.range)

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r}, {!r})'.format(self.__class__.__name__,
                                       self.domain, self.exponent)

    def __str__(self):
        """Return ``str(self)``."""
        return "x ** {}".format(self.exponent)


class InnerProductOperator(Operator):
    """Operator taking the inner product with a fixed space element.

    Implements::

        InnerProductOperator(y)(x) <==> y.inner(x)

    This is only applicable in inner product spaces.

    See Also
    --------
    DistOperator : Distance to a fixed space element.
    NormOperator : Vector space norm as operator.
    """

    def __init__(self, vector):
        """Initialize a new instance.

        Parameters
        ----------
        vector : `LinearSpaceElement`
            The element to take the inner product with.

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> x = r3.element([1, 2, 3])
        >>> op = InnerProductOperator(x)
        >>> op(r3.element([1, 2, 3]))
        14.0
        """
        super(InnerProductOperator, self).__init__(
            vector.space, vector.space.field, linear=True)
        self.__vector = vector

    @property
    def vector(self):
        """Element to take the inner product with."""
        return self.__vector

    def _call(self, x):
        """Return the inner product with ``x``."""
        return x.inner(self.vector)

    @property
    def adjoint(self):
        """Adjoint of this operator.

        Returns
        -------
        adjoint : `MultiplyOperator`
            The operator of multiplication with `vector`.

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> x = r3.element([1, 2, 3])
        >>> op = InnerProductOperator(x)
        >>> op.adjoint(2.0)
        rn(3).element([ 2.,  4.,  6.])
        """
        return MultiplyOperator(self.vector, self.vector.space.field)

    @property
    def T(self):
        """Fixed vector of this operator.

        Returns
        -------
        vector : `LinearSpaceElement`
            The fixed space element used in this inner product operator.

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> x = r3.element([1, 2, 3])
        >>> x.T
        InnerProductOperator(rn(3).element([ 1.,  2.,  3.]))
        >>> x.T.T
        rn(3).element([ 1.,  2.,  3.])
        """
        return self.vector

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r})'.format(self.__class__.__name__, self.vector)

    def __str__(self):
        """Return ``str(self)``."""
        return '{}.T'.format(self.vector)


class NormOperator(Operator):

    """Vector space norm as an operator.

    Implements::

        NormOperator()(x) <==> x.norm()

    This is only applicable in normed spaces, i.e., spaces implementing
    a ``norm`` method.

    See Also
    --------
    InnerProductOperator : Inner product with a fixed space element.
    DistOperator : Distance to a fixed space element.
    """

    def __init__(self, space):
        """Initialize a new instance.

        Parameters
        ----------
        space : `LinearSpace`
            Space to take the norm in.

        Examples
        --------
        >>> r2 = odl.rn(2)
        >>> op = NormOperator(r2)
        >>> op([3, 4])
        5.0
        """
        super(NormOperator, self).__init__(space, RealNumbers(), linear=False)

    def _call(self, x):
        """Return the norm of ``x``."""
        return x.norm()

    def derivative(self, point):
        r"""Derivative of this operator in ``point``.

            ``NormOperator().derivative(y)(x) == (y / y.norm()).inner(x)``

        This is only applicable in inner product spaces.

        Parameters
        ----------
        point : `domain` `element-like`
            Point in which to take the derivative.

        Returns
        -------
        derivative : `InnerProductOperator`

        Raises
        ------
        ValueError
            If ``point.norm() == 0``, in which case the derivative is not well
            defined in the Frechet sense.

        Notes
        -----
        The derivative cannot be written in a general sense except in Hilbert
        spaces, in which case it is given by

        .. math::
            (D \|\cdot\|)(y)(x) = \langle y / \|y\|, x \rangle

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> op = NormOperator(r3)
        >>> derivative = op.derivative([1, 0, 0])
        >>> derivative([1, 0, 0])
        1.0
        """
        point = self.domain.element(point)
        norm = point.norm()
        if norm == 0:
            raise ValueError('not differentiable in 0')

        return InnerProductOperator(point / norm)

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r})'.format(self.__class__.__name__, self.domain)

    def __str__(self):
        """Return ``str(self)``."""
        return '{}({})'.format(self.__class__.__name__, self.domain)


class DistOperator(Operator):

    """Operator taking the distance to a fixed space element.

    Implements::

        DistOperator(y)(x) == y.dist(x)

    This is only applicable in metric spaces, i.e., spaces implementing
    a ``dist`` method.

    See Also
    --------
    InnerProductOperator : Inner product with fixed space element.
    NormOperator : Vector space norm as an operator.
    """

    def __init__(self, vector):
        """Initialize a new instance.

        Parameters
        ----------
        vector : `LinearSpaceElement`
            Point to calculate the distance to.

        Examples
        --------
        >>> r2 = odl.rn(2)
        >>> x = r2.element([1, 1])
        >>> op = DistOperator(x)
        >>> op([4, 5])
        5.0
        """
        super(DistOperator, self).__init__(
            vector.space, RealNumbers(), linear=False)
        self.__vector = vector

    @property
    def vector(self):
        """Element to which to take the distance."""
        return self.__vector

    def _call(self, x):
        """Return the distance from ``self.vector`` to ``x``."""
        return self.vector.dist(x)

    def derivative(self, point):
        r"""The derivative operator.

            ``DistOperator(y).derivative(z)(x) ==
            ((y - z) / y.dist(z)).inner(x)``

        This is only applicable in inner product spaces.

        Parameters
        ----------
        x : `domain` `element-like`
            Point in which to take the derivative.

        Returns
        -------
        derivative : `InnerProductOperator`

        Raises
        ------
        ValueError
            If ``point == self.vector``, in which case the derivative is not
            well defined in the Frechet sense.

        Notes
        -----
        The derivative cannot be written in a general sense except in Hilbert
        spaces, in which case it is given by

        .. math::
            (D d(\cdot, y))(z)(x) = \langle (y-z) / d(y, z), x \rangle

        Examples
        --------
        >>> r2 = odl.rn(2)
        >>> x = r2.element([1, 1])
        >>> op = DistOperator(x)
        >>> derivative = op.derivative([2, 1])
        >>> derivative([1, 0])
        1.0
        """
        point = self.domain.element(point)
        diff = point - self.vector
        dist = self.vector.dist(point)
        if dist == 0:
            raise ValueError('not differentiable at the reference vector {!r}'
                             ''.format(self.vector))

        return InnerProductOperator(diff / dist)

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r})'.format(self.__class__.__name__, self.vector)

    def __str__(self):
        """Return ``str(self)``."""
        return '{}({})'.format(self.__class__.__name__, self.vector)


class ConstantOperator(Operator):

    """Operator that always returns the same value.

    Implements::

        ConstantOperator(y)(x) == y
    """

    def __init__(self, constant, domain=None, range=None):
        """Initialize a new instance.

        Parameters
        ----------
        constant : `LinearSpaceElement` or ``range`` `element-like`
            The constant space element to be returned. If ``range`` is not
            provided, ``constant`` must be a `LinearSpaceElement` since the
            operator range is then inferred from it.
        domain : `LinearSpace`, optional
            Domain of the operator. Default: ``vector.space``
        range : `LinearSpace`, optional
            Range of the operator. Default: ``vector.space``

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> x = r3.element([1, 2, 3])
        >>> op = ConstantOperator(x)
        >>> op(x, out=r3.element())
        rn(3).element([ 1.,  2.,  3.])
        """

        if ((domain is None or range is None) and
                not isinstance(constant, LinearSpaceElement)):
            raise TypeError('If either domain or range is unspecified '
                            '`constant` must be LinearSpaceVector, got '
                            '{!r}.'.format(constant))

        if domain is None:
            domain = constant.space
        if range is None:
            range = constant.space

        self.__constant = range.element(constant)
        linear = self.constant.norm() == 0
        super(ConstantOperator, self).__init__(domain, range, linear=linear)

    @property
    def constant(self):
        """Constant space element returned by this operator."""
        return self.__constant

    def _call(self, x, out=None):
        """Return the constant vector or assign it to ``out``."""
        if out is None:
            return self.range.element(copy(self.constant))
        else:
            out.assign(self.constant)

    @property
    def adjoint(self):
        """Adjoint of the operator.

        Only defined if the operator is the constant operator.
        """

    def derivative(self, point):
        """Derivative of this operator, always zero.

        Returns
        -------
        derivative : `ZeroOperator`

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> x = r3.element([1, 2, 3])
        >>> op = ConstantOperator(x)
        >>> deriv = op.derivative([1, 1, 1])
        >>> deriv([2, 2, 2])
        rn(3).element([ 0.,  0.,  0.])
        """
        return ZeroOperator(domain=self.domain, range=self.range)

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r})'.format(self.__class__.__name__, self.constant)

    def __str__(self):
        """Return ``str(self)``."""
        return "{}".format(self.constant)


class ZeroOperator(Operator):

    """Operator mapping each element to the zero element.

    Implements::

        ZeroOperator(space)(x) == space.zero()
    """

    def __init__(self, domain, range=None):
        """Initialize a new instance.

        Parameters
        ----------
        domain : `LinearSpace`
            Domain of the operator.
        range : `LinearSpace`, optional
            Range of the operator. Default: ``domain``

        Examples
        --------
        >>> op = odl.ZeroOperator(odl.rn(3))
        >>> op([1, 2, 3])
        rn(3).element([ 0.,  0.,  0.])

        Also works with domain != range:

        >>> op = odl.ZeroOperator(odl.rn(3), odl.cn(4))
        >>> op([1, 2, 3])
        cn(4).element([ 0.+0.j,  0.+0.j,  0.+0.j,  0.+0.j])
        """
        if range is None:
            range = domain

        super(ZeroOperator, self).__init__(domain, range, linear=True)

    def _call(self, x, out=None):
        """Return the zero vector or assign it to ``out``."""
        if self.domain == self.range:
            if out is None:
                out = 0 * x
            else:
                out.lincomb(0, x)
        else:
            result = self.range.zero()
            if out is None:
                out = result
            else:
                out.assign(result)
        return out

    @property
    def adjoint(self):
        """Adjoint of the operator.

        If ``self.domain == self.range`` the zero operator is self-adjoint,
        otherwise it is the `ZeroOperator` from `range` to `domain`.
        """
        return ZeroOperator(domain=self.range, range=self.domain)

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r})'.format(self.__class__.__name__, self.domain)

    def __str__(self):
        """Return ``str(self)``."""
        return '0'


class RealPart(Operator):

    """Operator that extracts the real part of a vector.

    Implements::

        RealPart(x) == x.real
    """

    def __init__(self, space):
        """Initialize a new instance.

        Parameters
        ----------
        space : `TensorSpace`
            Space in which the real part should be taken, needs to implement
            ``space.real_space``.

        Examples
        --------
        Take the real part of complex vector:

        >>> c3 = odl.cn(3)
        >>> op = RealPart(c3)
        >>> op([1 + 2j, 2, 3 - 1j])
        rn(3).element([ 1.,  2.,  3.])

        The operator is the identity on real spaces:

        >>> r3 = odl.rn(3)
        >>> op = RealPart(r3)
        >>> op([1, 2, 3])
        rn(3).element([ 1.,  2.,  3.])

        The operator also works on other `TensorSpace` spaces such as
        `DiscretizedSpace` spaces:

        >>> r3 = odl.uniform_discr(0, 1, 3, dtype=complex)
        >>> op = RealPart(r3)
        >>> op([1, 2, 3])
        uniform_discr(0.0, 1.0, 3).element([ 1.,  2.,  3.])
        """
        real_space = space.real_space
        self.space_is_real = (space == real_space)
        super(RealPart, self).__init__(space, real_space, linear=True)

    def _call(self, x):
        """Return ``self(x)``."""
        return x.real

    def derivative(self, x):
        r"""Return the derivative operator in the "C = R^2" sense.

        The returned operator (``self``) is the derivative of the
        operator variant where the complex domain is reinterpreted as
        a product of two real spaces.

        Parameters
        ----------
        x : `domain` element
            Point in which to take the derivative, has no effect.
        """
        return self

    @property
    def inverse(self):
        """Return the (pseudo-)inverse.

        Examples
        --------
        The inverse is its own inverse if its domain is real:

        >>> r3 = odl.rn(3)
        >>> op = RealPart(r3)
        >>> op.inverse(op([1, 2, 3]))
        rn(3).element([ 1.,  2.,  3.])

        This is not a true inverse, only a pseudoinverse, the complex part
        will by necessity be lost.

        >>> c3 = odl.cn(3)
        >>> op = RealPart(c3)
        >>> op.inverse(op([1 + 2j, 2, 3 - 1j]))
        cn(3).element([ 1.+0.j,  2.+0.j,  3.+0.j])
        """
        if self.space_is_real:
            return self
        else:
            return ComplexEmbedding(self.domain, scalar=1)

    @property
    def adjoint(self):
        r"""Return the (left) adjoint.

        Notes
        -----
        Due to technicalities of operators from a complex space into a real
        space, this does not satisfy the usual adjoint equation:

        .. math::
            \langle Ax, y \rangle = \langle x, A^*y \rangle

        Instead it is an adjoint in a weaker sense as follows:

        .. math::
            \langle AA^*x, y \rangle = \langle A^*x, A^*y \rangle

        Examples
        --------
        The adjoint satisfies the adjoint equation for real spaces:

        >>> r3 = odl.rn(3)
        >>> op = RealPart(r3)
        >>> x = op.domain.element([1, 2, 3])
        >>> y = op.range.element([3, 2, 1])
        >>> x.inner(op.adjoint(y)) == op(x).inner(y)
        True

        If the domain is complex, it only satisfies the weaker definition:

        >>> c3 = odl.cn(3)
        >>> op = RealPart(c3)
        >>> x = op.range.element([1, 2, 3])
        >>> y = op.range.element([3, 2, 1])
        >>> AtAxy = op(op.adjoint(x)).inner(y)
        >>> AtxAty = op.adjoint(x).inner(op.adjoint(y))
        >>> AtAxy == AtxAty
        True
        """
        if self.space_is_real:
            return self
        else:
            return ComplexEmbedding(self.domain, scalar=1)


class ImagPart(Operator):

    """Operator that extracts the imaginary part of a vector.

    Implements::

        ImagPart(x) == x.imag
    """

    def __init__(self, space):
        """Initialize a new instance.

        Parameters
        ----------
        space : `TensorSpace`
            Space in which the imaginary part should be taken, needs to
            implement ``space.real_space``.

        Examples
        --------
        Take the imaginary part of complex vector:

        >>> c3 = odl.cn(3)
        >>> op = ImagPart(c3)
        >>> op([1 + 2j, 2, 3 - 1j])
        rn(3).element([ 2.,  0., -1.])

        The operator is the zero operator on real spaces:

        >>> r3 = odl.rn(3)
        >>> op = ImagPart(r3)
        >>> op([1, 2, 3])
        rn(3).element([ 0.,  0.,  0.])
        """
        real_space = space.real_space
        self.space_is_real = (space == real_space)
        super(ImagPart, self).__init__(space, real_space, linear=True)

    def _call(self, x):
        """Return ``self(x)``."""
        return x.imag

    def derivative(self, x):
        r"""Return the derivative operator in the "C = R^2" sense.

        The returned operator (``self``) is the derivative of the
        operator variant where the complex domain is reinterpreted as
        a product of two real spaces.

        Parameters
        ----------
        x : `domain` element
            Point in which to take the derivative, has no effect.
        """
        return self

    @property
    def inverse(self):
        """Return the pseudoinverse.

        Examples
        --------
        The inverse is the zero operator if the domain is real:

        >>> r3 = odl.rn(3)
        >>> op = ImagPart(r3)
        >>> op.inverse(op([1, 2, 3]))
        rn(3).element([ 0.,  0.,  0.])

        This is not a true inverse, only a pseudoinverse, the real part
        will by necessity be lost.

        >>> c3 = odl.cn(3)
        >>> op = ImagPart(c3)
        >>> op.inverse(op([1 + 2j, 2, 3 - 1j]))
        cn(3).element([ 0.+2.j,  0.+0.j, -0.-1.j])
        """
        if self.space_is_real:
            return ZeroOperator(self.domain)
        else:
            return ComplexEmbedding(self.domain, scalar=1j)

    @property
    def adjoint(self):
        r"""Return the (left) adjoint.

        Notes
        -----
        Due to technicalities of operators from a complex space into a real
        space, this does not satisfy the usual adjoint equation:

        .. math::
            \langle Ax, y \rangle = \langle x, A^*y \rangle

        Instead it is an adjoint in a weaker sense as follows:

        .. math::
            \langle AA^*x, y \rangle = \langle A^*x, A^*y \rangle

        Examples
        --------
        The adjoint satisfies the adjoint equation for real spaces:

        >>> r3 = odl.rn(3)
        >>> op = ImagPart(r3)
        >>> x = op.domain.element([1, 2, 3])
        >>> y = op.range.element([3, 2, 1])
        >>> x.inner(op.adjoint(y)) == op(x).inner(y)
        True

        If the domain is complex, it only satisfies the weaker definition:

        >>> c3 = odl.cn(3)
        >>> op = ImagPart(c3)
        >>> x = op.range.element([1, 2, 3])
        >>> y = op.range.element([3, 2, 1])
        >>> AtAxy = op(op.adjoint(x)).inner(y)
        >>> AtxAty = op.adjoint(x).inner(op.adjoint(y))
        >>> AtAxy == AtxAty
        True
        """
        if self.space_is_real:
            return ZeroOperator(self.domain)
        else:
            return ComplexEmbedding(self.domain, scalar=1j)


class ComplexEmbedding(Operator):

    """Operator that embeds a vector into a complex space.

    Implements::

        ComplexEmbedding(space)(x) <==> space.complex_space.element(x)
    """

    def __init__(self, space, scalar=1.0):
        """Initialize a new instance.

        Parameters
        ----------
        space : `TensorSpace`
            Space that should be embedded into its complex counterpart.
            It must implement `TensorSpace.complex_space`.
        scalar : ``space.complex_space.field`` element, optional
            Scalar to be multiplied with incoming vectors in order
            to get the complex vector.

        Examples
        --------
        Embed real vector into complex space:

        >>> r3 = odl.rn(3)
        >>> op = ComplexEmbedding(r3)
        >>> op([1, 2, 3])
        cn(3).element([ 1.+0.j,  2.+0.j,  3.+0.j])

        Embed real vector as imaginary part into complex space:

        >>> op = ComplexEmbedding(r3, scalar=1j)
        >>> op([1, 2, 3])
        cn(3).element([ 0.+1.j,  0.+2.j,  0.+3.j])

        On complex spaces the operator is the same as simple multiplication by
        scalar:

        >>> c3 = odl.cn(3)
        >>> op = ComplexEmbedding(c3, scalar=1 + 2j)
        >>> op([1 + 1j, 2 + 2j, 3 + 3j])
        cn(3).element([-1.+3.j, -2.+6.j, -3.+9.j])
        """
        complex_space = space.complex_space
        self.scalar = complex_space.field.element(scalar)
        super(ComplexEmbedding, self).__init__(
            space, complex_space, linear=True)

    def _call(self, x, out):
        """Return ``self(x)``."""
        if self.domain.is_real:
            # Real domain, multiply separately
            out.real = self.scalar.real * x
            out.imag = self.scalar.imag * x
        else:
            # Complex domain
            out.lincomb(self.scalar, x)

    @property
    def inverse(self):
        """Return the (left) inverse.

        If the domain is a real space, this is not a true inverse,
        only a (left) inverse.

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> op = ComplexEmbedding(r3, scalar=1)
        >>> op.inverse(op([1, 2, 4]))
        rn(3).element([ 1.,  2.,  4.])
        """
        if self.domain.is_real:
            # Real domain
            # Optimizations for simple cases.
            if self.scalar.real == self.scalar:
                return (1 / self.scalar.real) * RealPart(self.range)
            elif 1j * self.scalar.imag == self.scalar:
                return (1 / self.scalar.imag) * ImagPart(self.range)
            else:
                # General case
                inv_scalar = (1 / self.scalar).conjugate()
                return ((inv_scalar.real) * RealPart(self.range) +
                        (inv_scalar.imag) * ImagPart(self.range))
        else:
            # Complex domain
            return ComplexEmbedding(self.range, self.scalar.conjugate())

    @property
    def adjoint(self):
        r"""Return the (right) adjoint.

        Notes
        -----
        Due to technicalities of operators from a real space into a complex
        space, this does not satisfy the usual adjoint equation:

        .. math::
            \langle Ax, y \rangle = \langle x, A^*y \rangle

        Instead it is an adjoint in a weaker sense as follows:

        .. math::
            \langle A^*Ax, y \rangle = \langle Ax, Ay \rangle

        Examples
        --------
        The adjoint satisfies the adjoint equation for complex spaces

        >>> c3 = odl.cn(3)
        >>> op = ComplexEmbedding(c3, scalar=1j)
        >>> x = c3.element([1 + 1j, 2 + 2j, 3 + 3j])
        >>> y = c3.element([3 + 1j, 2 + 2j, 3 + 1j])
        >>> Axy = op(x).inner(y)
        >>> xAty = x.inner(op.adjoint(y))
        >>> Axy == xAty
        True

        For real domains, it only satisfies the (right) adjoint equation

        >>> r3 = odl.rn(3)
        >>> op = ComplexEmbedding(r3, scalar=1j)
        >>> x = r3.element([1, 2, 3])
        >>> y = r3.element([3, 2, 3])
        >>> AtAxy = op.adjoint(op(x)).inner(y)
        >>> AxAy = op(x).inner(op(y))
        >>> AtAxy == AxAy
        True
        """
        if self.domain.is_real:
            # Real domain
            # Optimizations for simple cases.
            if self.scalar.real == self.scalar:
                return self.scalar.real * RealPart(self.range)
            elif 1j * self.scalar.imag == self.scalar:
                return self.scalar.imag * ImagPart(self.range)
            else:
                # General case
                return (self.scalar.real * RealPart(self.range) +
                        self.scalar.imag * ImagPart(self.range))
        else:
            # Complex domain
            return ComplexEmbedding(self.range, self.scalar.conjugate())


class ComplexModulus(Operator):

    """Operator that computes the modulus (absolute value) of a vector."""

    def __init__(self, space):
        """Initialize a new instance.

        Parameters
        ----------
        space : `TensorSpace`
            Space in which the modulus should be taken, needs to implement
            ``space.real_space``.

        Examples
        --------
        Take the modulus of a complex vector:

        >>> c2 = odl.cn(2)
        >>> op = odl.ComplexModulus(c2)
        >>> op([3 + 4j, 2])
        rn(2).element([ 5.,  2.])

        The operator is the absolute value on real spaces:

        >>> r2 = odl.rn(2)
        >>> op = odl.ComplexModulus(r2)
        >>> op([1, -2])
        rn(2).element([ 1.,  2.])

        The operator also works on other `TensorSpace`'s such as
        `DiscretizedSpace`:

        >>> space = odl.uniform_discr(0, 1, 2, dtype=complex)
        >>> op = odl.ComplexModulus(space)
        >>> op([3 + 4j, 2])
        uniform_discr(0.0, 1.0, 2).element([ 5.,  2.])
        """
        real_space = space.real_space
        super(ComplexModulus, self).__init__(space, real_space, linear=False)

    def _call(self, x):
        """Return ``self(x)``."""
        return sqrt(x.real ** 2 + x.imag ** 2)

    def derivative(self, x):
        r"""Return the derivative operator in the "C = R^2" sense.

        The returned operator (``self``) is the derivative of the
        operator variant where the complex domain is reinterpreted as
        a product of two real spaces::

            M'(x) = y --> ((Re(x) * Re(y) + Im(x) * Im(y)) /
                           sqrt(Re(x)**2 + Re(y) ** 2))

        Parameters
        ----------
        x : `domain` element
            Point in which to take the derivative.

        Examples
        --------
        >>> c2 = odl.cn(2)
        >>> op = odl.ComplexModulus(c2)
        >>> op([3 + 4j, 2])
        rn(2).element([ 5.,  2.])
        >>> deriv = op.derivative([3 + 4j, 2])
        >>> deriv.domain
        cn(2)
        >>> deriv.range
        rn(2)
        >>> deriv([2 + 1j, 4j])  # [(3*2 + 4*1) / 5, (2*0 + 0*4) / 2]
        rn(2).element([ 2.,  0.])

        Notes
        -----
        The derivative of the complex modulus

        .. math::
            &M: X(\mathbb{C}) \to X(\mathbb{R}), \\
            &M(x) = \sqrt{\Re(x)^2 + \Im(x)^2},

        with :math:`X(\mathbb{F}) = \mathbb{F}^n` or
        :math:`L^2(\Omega, \mathbb{F})`, is given as

        .. math::
            &M'(x): X(\mathbb{C}) \to X(\mathbb{R}), \\
            &M'(x)(y) = \frac{\Re(x)\,\Re(y) + \Im(x)\,\Im(y)}{M(x)}.

        It is linear when identifying :math:`\mathbb{C}` with
        :math:`\mathbb{R}^2`, but not complex-linear.
        """
        op = self
        x = self.domain.element(x)

        class ComplexModulusDerivative(Operator):

            """Derivative of the complex modulus operator."""

            def _call(self, y, out):
                """Return ``self(y)``."""
                out[:] = x.real * y.real
                out += x.imag * y.imag
                out /= op(x)
                return out

            @property
            def adjoint(self):
                r"""Adjoint in the "C = R^2" sense.

                Examples
                --------
                Adjoint of the derivative:

                >>> c2 = odl.cn(2)
                >>> op = odl.ComplexModulus(c2)
                >>> op([3 + 4j, 2])
                rn(2).element([ 5.,  2.])
                >>> deriv = op.derivative([3 + 4j, 2])
                >>> adj = deriv.adjoint
                >>> adj.domain
                rn(2)
                >>> adj.range
                cn(2)
                >>> adj([5, 5])  # [5*(3 + 4j)/5, 5*2/2]
                cn(2).element([ 3.+4.j,  5.+0.j])

                Adjointness only holds in the weaker sense that inner products
                are the same when testing with vectors from the real space, but
                not when testing complex vectors:

                >>> y1 = deriv.range.element([5, 5])
                >>> y2 = deriv.range.element([1, 2])
                >>> adj(y1).inner(adj(y2))  # <M^* y1, M^* y2>
                (15+0j)
                >>> deriv(adj(y1)).inner(y2)  # <M M^* y1, y2>
                15.0
                >>> x1 = deriv.domain.element([6 + 3j, 2j])
                >>> x2 = deriv.domain.element([5, 10 + 4j])
                >>> deriv(x1).inner(deriv(x2))  # <M x1, M x2>
                18.0
                >>> adj(deriv(x1)).inner(x2)  # <M^* M x1, x2>
                (18+24j)

                Notes
                -----
                The complex modulus derivative is given by

                .. math::
                    &M'(x): X(\mathbb{C}) \to X(\mathbb{R}), \\
                    &M'(x)(y) = \frac{\Re(x)\,\Re(y) + \Im(x)\,\Im(y)}{M(x)}.

                Thus, its adjoint can (formally) be identified as

                .. math::
                    &M'(x)^*: X(\mathbb{R}) \to X(\mathbb{C}), \\
                    &M'(x)^*(u) = \frac{(\Re(x)\,u,\ \Im(x)\,u}{M(x)}.

                The operator :math:`A = M'(x)` has the weak adjointness
                property

                .. math::
                    \langle A^* y_1,\ A^* y_2 \rangle_{X(\mathbb{C})} =
                    \langle AA^* y_1,\ y_2 \rangle_{X(\mathbb{R})},

                but in general,

                .. math::
                    \langle A x,\ y \rangle_{X(\mathbb{R})} \neq
                    \langle x,\ A^* y \rangle_{X(\mathbb{C})},

                in particular

                .. math::
                    \langle A x_1,\ A x_2 \rangle_{X(\mathbb{R})} \neq
                    \langle A^*A x_1,\ x_2 \rangle_{X(\mathbb{C})}.
                """
                deriv = self

                class ComplexModulusDerivativeAdjoint(Operator):

                    def _call(self, u, out):
                        """Implement ``self(u, out)``."""
                        out.assign(x)
                        tmp = u / op(x)
                        out.real *= tmp
                        out.imag *= tmp
                        return out

                    @property
                    def adjoint(self):
                        """Adjoint in the "C = R^2" sense."""
                        return deriv

                return ComplexModulusDerivativeAdjoint(
                    deriv.range, deriv.domain, linear=True
                )

        return ComplexModulusDerivative(op.domain, op.range, linear=True)


class ComplexModulusSquared(Operator):

    """Operator that computes the squared complex modulus (absolute value)."""

    def __init__(self, space):
        """Initialize a new instance.

        Parameters
        ----------
        space : `TensorSpace`
            Space in which the modulus should be taken, needs to implement
            ``space.real_space``.

        Examples
        --------
        Take the squared modulus of a complex vector:

        >>> c2 = odl.cn(2)
        >>> op = odl.ComplexModulusSquared(c2)
        >>> op([3 + 4j, 2])
        rn(2).element([ 25.,   4.])

        On a real space, this is the same as squaring:

        >>> r2 = odl.rn(2)
        >>> op = odl.ComplexModulusSquared(r2)
        >>> op([1, -2])
        rn(2).element([ 1.,  4.])

        The operator also works on other `TensorSpace`'s such as
        `DiscretizedSpace`:

        >>> space = odl.uniform_discr(0, 1, 2, dtype=complex)
        >>> op = odl.ComplexModulusSquared(space)
        >>> op([3 + 4j, 2])
        uniform_discr(0.0, 1.0, 2).element([ 25.,   4.])
        """
        real_space = space.real_space
        super(ComplexModulusSquared, self).__init__(
            space, real_space, linear=False)

    def _call(self, x):
        """Return ``self(x)``."""
        return x.real ** 2 + x.imag ** 2

    def derivative(self, x):
        r"""Return the derivative operator in the "C = R^2" sense.

        The returned operator (``self``) is the derivative of the
        operator variant where the complex domain is reinterpreted as
        a product of two real spaces.

        Parameters
        ----------
        x : `domain` element
            Point in which to take the derivative.

        Examples
        --------
        >>> c2 = odl.cn(2)
        >>> op = odl.ComplexModulusSquared(c2)
        >>> op([3 + 4j, 2])
        rn(2).element([ 25.,   4.])
        >>> deriv = op.derivative([3 + 4j, 2])
        >>> deriv.domain
        cn(2)
        >>> deriv.range
        rn(2)
        >>> deriv([2 + 1j, 4j])  # [(3*2 + 4*1) * 2, (2*0 + 0*4) * 2]
        rn(2).element([ 20.,   0.])

        Notes
        -----
        The derivative of the squared complex modulus

        .. math::
            &S: X(\mathbb{C}) \to X(\mathbb{R}), \\
            &S(x) = 2\Re(x)^2 + \Im(x)^2,

        with :math:`X(\mathbb{F}) = \mathbb{F}^n` or
        :math:`L^2(\Omega, \mathbb{F})`, is given as

        .. math::
            &S'(x): X(\mathbb{C}) \to X(\mathbb{R}), \\
            &S'(x)(y) = 2(\Re(x)\,\Re(y) + \Im(x)\,\Im(y)).

        It is linear when identifying :math:`\mathbb{C}` with
        :math:`\mathbb{R}^2`, but not complex-linear.
        """
        op = self
        x = self.domain.element(x)

        class ComplexModulusSquaredDerivative(Operator):

            """Derivative of the squared complex modulus operator."""

            def _call(self, y, out):
                """Return ``self(y)``."""
                x.real.multiply(y.real, out=out)
                out += x.imag * y.imag
                out *= 2
                return out

            @property
            def adjoint(self):
                r"""Adjoint in the "C = R^2" sense.

                Adjoint of the derivative:

                Examples
                --------
                >>> c2 = odl.cn(2)
                >>> op = odl.ComplexModulusSquared(c2)
                >>> deriv = op.derivative([3 + 4j, 2])
                >>> adj = deriv.adjoint
                >>> adj.domain
                rn(2)
                >>> adj.range
                cn(2)
                >>> adj([2, 1])  # 2 * [2*(3 + 4j), 1*2]
                cn(2).element([ 12.+16.j,   4. +0.j])

                Adjointness only holds in the weaker sense that inner products
                are the same when testing with vectors from the real space, but
                not when testing complex vectors:

                >>> y1 = deriv.range.element([1, 1])
                >>> y2 = deriv.range.element([1, -1])
                >>> adj(y1).inner(adj(y2))  # <M^* y1, M^* y2>
                (84+0j)
                >>> deriv(adj(y1)).inner(y2)  # <M M^* y1, y2>
                84.0
                >>> x1 = deriv.domain.element([1j, 1j])
                >>> x2 = deriv.domain.element([1 + 1j, 1j])
                >>> deriv(x1).inner(deriv(x2))  # <M x1, M x2>
                112.0
                >>> adj(deriv(x1)).inner(x2)  # <M^* M x1, x2>
                (112+16j)

                Notes
                -----
                The squared complex modulus derivative is given by

                .. math::
                    &S'(x): X(\mathbb{C}) \to X(\mathbb{R}), \\
                    &S'(x)(y) = 2(\Re(x)\,\Re(y) + \Im(x)\,\Im(y)).

                Thus, its adjoint can (formally) be identified as

                .. math::
                    &S'(x)^*: X(\mathbb{R}) \to X(\mathbb{C}), \\
                    &S'(x)^*(u) = 2(\Re(x)\,u,\ \Im(x)\,u).

                The operator :math:`A = S'(x)` has the weak adjointness
                property

                .. math::
                    \langle A^* y_1,\ A^* y_2 \rangle_{X(\mathbb{C})} =
                    \langle AA^* y_1,\ y_2 \rangle_{X(\mathbb{R})},

                but in general,

                .. math::
                    \langle A x,\ y \rangle_{X(\mathbb{R})} \neq
                    \langle x,\ A^* y \rangle_{X(\mathbb{C})},

                in particular

                .. math::
                    \langle A x_1,\ A x_2 \rangle_{X(\mathbb{R})} \neq
                    \langle A^*A x_1,\ x_2 \rangle_{X(\mathbb{C})}.
                """
                deriv = self

                class ComplexModulusSquaredDerivAdj(Operator):

                    def _call(self, u, out):
                        """Implement ``self(u, out)``."""
                        out.assign(x)
                        out.real *= u
                        out.imag *= u
                        out *= 2
                        return out

                    @property
                    def adjoint(self):
                        """Adjoint in the "C = R^2" sense."""
                        return deriv

                return ComplexModulusSquaredDerivAdj(
                    deriv.range, deriv.domain, linear=True
                )

        return ComplexModulusSquaredDerivative(
            op.domain, op.range, linear=True
        )


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.
#
# The data is licensed under a
# Creative Commons Attribution 4.0 International License.
#
# You should have received a copy of the license along with this
# work. If not, see <http://creativecommons.org/licenses/by/4.0/>.

"""Tomographic datasets from the Finish Inverse Problems Society (FIPS)."""

from __future__ import division
import numpy as np
from odl.contrib.datasets.util import get_data
from odl.core.discr import uniform_partition
from odl.applications.tomo import FanBeamGeometry


__all__ = ('walnut_data', 'walnut_geometry',
           'lotus_root_data', 'lotus_root_geometry')


DATA_SUBSET = 'ct_fips'


def walnut_data():
    """Tomographic X-ray data of a walnut.

    Notes
    -----
    See the article `Tomographic X-ray data of a walnut`_ for further
    information.

    See Also
    --------
    walnut_geometry

    References
    ----------
    .. _Tomographic X-ray data of a walnut: https://arxiv.org/abs/1502.04064
    """
    # TODO: Store data in some ODL controlled url
    url = 'https://zenodo.org/records/1254206/files/FullSizeSinograms.mat'
    dct = get_data('walnut.mat', subset=DATA_SUBSET, url=url)

    # Change axes to match ODL definitions
    data = np.swapaxes(dct['sinogram1200'], 0, 1)[::-1, ::-1]
    data = data.astype('float')

    # Very crude gain normalization
    data = -np.log(data / np.max(data, axis=1)[:, None])
    return data


def walnut_geometry():
    """Tomographic geometry for the walnut dataset.

    Notes
    -----
    See the article `Tomographic X-ray data of a walnut`_ for further
    information.

    See Also
    --------
    walnut_data

    References
    ----------
    .. _Tomographic X-ray data of a walnut: https://arxiv.org/abs/1502.04064
    """
    # To get the same rotation as in the reference article
    a_offset = -np.pi / 2
    apart = uniform_partition(a_offset, a_offset + 2 * np.pi, 1200)

    # TODO: Find exact value, determined experimentally
    d_offset = -0.279
    dpart = uniform_partition(d_offset - 57.4, d_offset + 57.4, 2296)

    geometry = FanBeamGeometry(apart, dpart,
                               src_radius=110, det_radius=190)

    return geometry


def lotus_root_data():
    """Tomographic X-ray data of a lotus root.

    Notes
    -----
    See the article `Tomographic X-ray data of a lotus root filled with
    attenuating objects`_ for further information.

    See Also
    --------
    lotus_root_geometry

    References
    ----------
    .. _Tomographic X-ray data of a lotus root filled with attenuating objects:
       https://arxiv.org/abs/1609.07299
    """
    # TODO: Store data in some ODL controlled url
    url = 'https://zenodo.org/records/1254204/files/sinogram.mat?download=1'
    dct = get_data('lotus_root.mat', subset=DATA_SUBSET, url=url)

    # Change axes to match ODL definitions
    data = np.swapaxes(dct['sinogram'], 0, 1)[:, :]
    data = data.astype('float')

    return data


def lotus_root_geometry():
    """Tomographic geometry for the lotus root dataset.

    Notes
    -----
    See the article `Tomographic X-ray data of a lotus root filled with
    attenuating objects`_ for further information.

    See Also
    --------
    lotus_root_geometry

    References
    ----------
    .. _Tomographic X-ray data of a lotus root filled with attenuating objects:
       https://arxiv.org/abs/1609.07299
    """
    # To get the same rotation as in the reference article
    a_offset = np.pi / 2
    apart = uniform_partition(a_offset,
                              a_offset + 2 * np.pi * 366. / 360.,
                              366)

    # TODO: Find exact value, determined experimentally
    d_offset = 0.35
    dpart = uniform_partition(d_offset - 60, d_offset + 60, 2240)

    geometry = FanBeamGeometry(apart, dpart,
                               src_radius=540, det_radius=90)

    return geometry


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.
#
# The images are (c) by Matthias Joachim Ehrhardt, University of Cambridge.
#
# The images are licensed under a
# Creative Commons Attribution 4.0 International License.
#
# You should have received a copy of the license along with this
# work. If not, see <http://creativecommons.org/licenses/by/4.0/>.


"""Images provided by the University of Cambridge."""

import numpy as np
import skimage.transform
from odl.contrib.datasets.util import get_data

__all__ = ('brain_phantom', 'resolution_phantom', 'building', 'rings',
           'blurring_kernel')


DATA_SUBSET = 'images_cambridge'
URL_CAM = 'https://raw.github.com/mehrhardt/spdhg/master/data/'


def convert(image, shape, gray=False, dtype='float64', normalize='max'):
    """Convert image to standardized format.

    Several properties of the input image may be changed including the shape,
    data type and maximal value of the image. In addition, this function may
    convert the image into an ODL object and/or a gray scale image.
    """

    image = image.astype(dtype)

    if gray:
        image[..., 0] *= 0.2126
        image[..., 1] *= 0.7152
        image[..., 2] *= 0.0722
        image = np.sum(image, axis=2)

    if shape is not None:
        image = skimage.transform.resize(image, shape, mode='constant')
        image = image.astype(dtype)

    if normalize == 'max':
        image /= image.max()
    elif normalize == 'sum':
        image /= image.sum()
    else:
        assert False

    return image


def brain_phantom(shape=None):
    """Brain phantom for FDG PET simulations.

    Returns
    -------
    An image with the following properties:
        image type: gray scales
        shape: [1024, 1024] (if not specified by `size`)
        scale: [0, 1]
        type: float64
    """
    # TODO: Store data in some ODL controlled url
    name = 'PET_phantom.mat'
    url = URL_CAM + name
    dct = get_data(name, subset=DATA_SUBSET, url=url)
    im = np.rot90(dct['im'], k=3)

    return convert(im, shape)


def resolution_phantom(shape=None):
    """Resolution phantom for tomographic simulations.

    Returns
    -------
    An image with the following properties:
        image type: gray scales
        shape: [1024, 1024] (if not specified by `size`)
        scale: [0, 1]
        type: float64
    """
    # TODO: Store data in some ODL controlled url
    # TODO: This can be also done with ODL's ellipse_phantom
    name = 'phantom_resolution.mat'
    url = URL_CAM + name
    dct = get_data(name, subset=DATA_SUBSET, url=url)
    im = np.rot90(dct['im'], k=3)

    return convert(im, shape)


def building(shape=None, gray=False):
    """Photo of the Centre for Mathematical Sciences in Cambridge.

    Returns
    -------
    An image with the following properties:
        image type: color (or gray scales if `gray=True`)
        size: [442, 331] (if not specified by `size`)
        scale: [0, 1]
        type: float64
    """
    # TODO: Store data in some ODL controlled url
    name = 'cms.mat'
    url = URL_CAM + name
    dct = get_data(name, subset=DATA_SUBSET, url=url)
    im = np.rot90(dct['im'], k=3)

    return convert(im, shape, gray=gray)


def rings(shape=None, gray=False):
    """Photo of married couple holding hands.

    Returns
    -------
    An image with the following properties:
        image type: color (or gray scales if `gray=True`)
        size: [3264, 2448] (if not specified by `size`)
        scale: [0, 1]
        type: float64
    """
    # TODO: Store data in some ODL controlled url
    name = 'rings.mat'
    url = URL_CAM + name
    dct = get_data(name, subset=DATA_SUBSET, url=url)
    im = np.rot90(dct['im'], k=2)

    return convert(im, shape, gray=gray)


def blurring_kernel(shape=None):
    """Blurring kernel for convolution simulations.

    The kernel is scaled to sum to one.

    Returns
    -------
    An image with the following properties:
        image type: gray scales
        size: [100, 100] (if not specified by `size`)
        scale: [0, 1]
        type: float64

    """
    # TODO: Store data in some ODL controlled url
    name = 'motionblur.mat'
    url = URL_CAM + name
    dct = get_data(name, subset=DATA_SUBSET, url=url)

    return convert(255 - dct['im'], shape, normalize='sum')


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Utilities mainly for internal use."""

from __future__ import absolute_import

from .graphics import *
from .normalize import *
from .npy_compat import *
from .numerics import *
from .testutils import *
from .utility import *
from .vectorization import *
from .sparse import *
from .scipy_compatibility import *
from .dtype_utils import *
from .print_utils import *

__all__ = ()
__all__ += graphics.__all__
__all__ += normalize.__all__
__all__ += npy_compat.__all__
__all__ += numerics.__all__
__all__ += testutils.__all__
__all__ += utility.__all__
__all__ += vectorization.__all__
__all__ += sparse.__all__
__all__ += scipy_compatibility.__all__
__all__ += dtype_utils.__all__
__all__ += print_utils.__all__



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Discretization-related functionality like grids and discrete spaces."""

from __future__ import absolute_import

from . import discr_utils
from .diff_ops import *
from .discr_ops import *
from .discr_space import *
from .grid import *
from .partition import *

__all__ = ()
__all__ += diff_ops.__all__
__all__ += discr_ops.__all__
__all__ += discr_space.__all__
__all__ += grid.__all__
__all__ += partition.__all__



# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""
Comparisons functions
    -> Utility functions expected by the python array API: `all` and `any` 
    -> Convenience functions that work in both backends: `allclose` and `isclose`
    -> Convenient composition of two functions: `all_equal`

Args:
    x (Number, LinearSpaceElement): Left-hand operand
    y (Number, LinearSpaceElement): Right-hand operand

Returns:
    x (bool | array-like of bools): Output of the comparison

Notes:
    1) These functions do not return ODL objects
"""

from .utils import get_array_and_backend
from numbers import Number
import numpy as np

__all__ = (
    "all",    
    "allclose",
    "odl_all_equal",
    "any",
    "isclose"
)


def _helper(x, fname, **kwargs):
    """
    Examples
    >>> space = odl.rn(3)
    >>> e1 = space.element((1,2,3))
    >>> odl.isclose(e1, space.element([1,2,3]))
    array([ True,  True,  True], dtype=bool)
    >>> odl.isclose(e1, space.element([1.5,2,3.2]))
    array([False,  True,  False], dtype=bool)
    >>> odl.allclose(e1, space.element([1,2,3]))
    True
    >>> odl.allclose(e1, space.element([2,2,2]))
    False
    >>> odl.all(odl.isclose(e1, space.element([1,2,3])))
    True
    >>> odl.any(odl.isclose(e1, space.element([1.5,2,3.2])))
    True
    >>> odl.all(e1 == [1,2,3])
    Traceback (most recent call last):
    ValueError: The left hand operand is a python Number of type <class 'bool'> and no right hand arguments were provided.
    """
    if isinstance(x, Number):        
        if 'y' in kwargs:
            y = kwargs.pop('y')
            if isinstance(y, Number):
                fn = getattr(np, fname)
            else:
                y, backend_y = get_array_and_backend(y)
                fn = getattr(backend_y.array_namespace, fname)
            # Devilish pytorch call for eq
            # https://docs.pytorch.org/docs/2.7/generated/torch.eq.html
            if fname == 'equal':
                return fn(y, x, **kwargs)
            else:
                return fn(x, y, **kwargs)
        else: 
            raise ValueError(f"The left hand operand is a python Number of type {type(x)} and no right hand arguments were provided.")
        
    x, backend_x = get_array_and_backend(x)
    fn = getattr(backend_x.array_namespace, fname)
    if 'y' in kwargs:
        y = kwargs.pop('y')
        if isinstance(y, Number):
            pass
        else:
            y, backend_y = get_array_and_backend(y)
            assert backend_x == backend_y, f"Two different backends {backend_x.impl} and {backend_y.impl} were provided, This operation is not supported by odl functions. Please ensure that your objects have the same implementation."
        return fn(x, y, **kwargs)
    else:
        return fn(x, **kwargs)

def all(x):
    """
    Test whether all array elements along a given axis evaluate to True.
    """
    return _helper(x, 'all')

def allclose(x, y, rtol=1e-05, atol=1e-08, equal_nan=False):
    """
    Returns True if two arrays are element-wise equal within a tolerance.
    Note: This is not a Python Array API method, but it happens to work in Numpy and Pytorch.
    """
    return _helper(x, 'allclose', y=y, rtol=rtol, atol=atol, equal_nan=equal_nan)

def odl_all_equal(x, y):
    """
    Test whether all array elements along a given axis evaluate to True.
    Note: This is not a Python Array API method, but a composition for convenience.
    It requires both sides of the comparison to use the same implementation, like
    other array-API functions, to avoid inefficient copying / restructuring.
    For a more flexible equality check useful for testing purposes, consider
    `all_equal` from `odl.testutils`.
    """
    return _helper(_helper(x, 'equal', y=y), 'all')

def any(x):
    """
    Test whether any array element along a given axis evaluates to True.
    """
    return _helper(x, 'any')

def isclose(x, y, rtol=1e-05, atol=1e-08, equal_nan=False):
    """
    Returns a boolean array where two arrays are element-wise equal within a tolerance.
    Note: This is not a Python Array API method, but it happens to work in Numpy and Pytorch.
    """
    return _helper(x, 'isclose', y=y, rtol=rtol, atol=atol, equal_nan=equal_nan)




# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Stochastic Primal-Dual Hybrid Gradient (SPDHG) algorithms"""

from __future__ import print_function, division
import numpy as np
import odl

__all__ = ('pdhg', 'spdhg', 'pa_spdhg', 'spdhg_generic', 'da_spdhg',
           'spdhg_pesquet')


def pdhg(x, f, g, A, tau, sigma, niter, **kwargs):
    """Computes a saddle point with PDHG.

    This algorithm is the same as "algorithm 1" in [CP2011a] but with
    extrapolation on the dual variable.


    Parameters
    ----------
    x : primal variable
        This variable is both input and output of the method.
    f : function
        Functional Y -> IR_infty that has a convex conjugate with a
        proximal operator, i.e. f.convex_conj.proximal(sigma) : Y -> Y.
    g : function
        Functional X -> IR_infty that has a proximal operator, i.e.
        g.proximal(tau) : X -> X.
    A : function
        Operator A : X -> Y that possesses an adjoint: A.adjoint
    tau : scalar / vector / matrix
        Step size for primal variable. Note that the proximal operator of g
        has to be well-defined for this input.
    sigma : scalar
        Scalar / vector / matrix used as step size for dual variable. Note that
        the proximal operator related to f (see above) has to be well-defined
        for this input.
    niter : int
        Number of iterations

    Other Parameters
    ----------------
    y: dual variable
        Dual variable is part of a product space
    z: variable
        Adjoint of dual variable, z = A^* y.
    theta : scalar
        Extrapolation factor.
    callback : callable
        Function called with the current iterate after each iteration.

    References
    ----------
    [CP2011a] Chambolle, A and Pock, T. *A First-Order
    Primal-Dual Algorithm for Convex Problems with Applications to
    Imaging*. Journal of Mathematical Imaging and Vision, 40 (2011),
    pp 120-145.
    """

    def fun_select(k):
        return [0]

    f = odl.functionals.SeparableSum(f)
    A = odl.BroadcastOperator(A, 1)

    # Dual variable
    y = kwargs.pop('y', None)
    if y is None:
        y_new = None
    else:
        y_new = A.range.element([y])

    spdhg_generic(x, f, g, A, tau, [sigma], niter, fun_select, y=y_new,
                  **kwargs)

    if y is not None:
        y.assign(y_new[0])


def spdhg(x, f, g, A, tau, sigma, niter, **kwargs):
    r"""Computes a saddle point with a stochastic PDHG.

    This means, a solution (x*, y*), y* = (y*_1, ..., y*_n) such that

    (x*, y*) in arg min_x max_y sum_i=1^n <y_i, A_i> - f*[i](y_i) + g(x)

    where g : X -> IR_infty and f[i] : Y[i] -> IR_infty are convex, l.s.c. and
    proper functionals. For this algorithm, they all may be non-smooth and no
    strong convexity is assumed.

    Parameters
    ----------
    x : primal variable
        This variable is both input and output of the method.
    f : functions
        Functionals Y[i] -> IR_infty that all have a convex conjugate with a
        proximal operator, i.e.
        f[i].convex_conj.proximal(sigma[i]) : Y[i] -> Y[i].
    g : function
        Functional X -> IR_infty that has a proximal operator, i.e.
        g.proximal(tau) : X -> X.
    A : functions
        Operators A[i] : X -> Y[i] that possess adjoints: A[i].adjoint
    tau : scalar / vector / matrix
        Step size for primal variable. Note that the proximal operator of g
        has to be well-defined for this input.
    sigma : scalar
        Scalar / vector / matrix used as step size for dual variable. Note that
        the proximal operator related to f (see above) has to be well-defined
        for this input.
    niter : int
        Number of iterations

    Other Parameters
    ----------------
    y : dual variable
        Dual variable is part of a product space. By default equals 0.
    z : variable
        Adjoint of dual variable, z = A^* y. By default equals 0 if y = 0.
    theta : scalar
        Global extrapolation factor.
    prob: list
        List of probabilities that an index i is selected each iteration. By
        default this is uniform serial sampling, p_i = 1/n.
    fun_select : function
        Function that selects blocks at every iteration IN -> {1,...,n}. By
        default this is serial sampling, fun_select(k) selects an index
        i \in {1,...,n} with probability p_i.
    callback : callable
        Function called with the current iterate after each iteration.

    References
    ----------
    [CERS2017] A. Chambolle, M. J. Ehrhardt, P. Richtarik and C.-B. Schoenlieb,
    *Stochastic Primal-Dual Hybrid Gradient Algorithm with Arbitrary Sampling
    and Imaging Applications*. ArXiv: http://arxiv.org/abs/1706.04957 (2017).

    [E+2017] M. J. Ehrhardt, P. J. Markiewicz, P. Richtarik, J. Schott,
    A. Chambolle and C.-B. Schoenlieb, *Faster PET reconstruction with a
    stochastic primal-dual hybrid gradient method*. Wavelets and Sparsity XVII,
    58 (2017) http://doi.org/10.1117/12.2272946.
    """

    # Probabilities
    prob = kwargs.pop('prob', None)
    if prob is None:
        prob = [1 / len(A)] * len(A)

    # Selection function
    fun_select = kwargs.pop('fun_select', None)
    if fun_select is None:
        def fun_select(x):
            return [int(np.random.choice(len(A), 1, p=prob))]

    # Dual variable
    y = kwargs.pop('y', None)

    extra = [1 / p for p in prob]

    spdhg_generic(x, f, g, A, tau, sigma, niter, fun_select=fun_select, y=y,
                  extra=extra, **kwargs)


def pa_spdhg(x, f, g, A, tau, sigma, niter, mu_g, **kwargs):
    r"""Computes a saddle point with a stochastic PDHG and primal acceleration.

    Next to other standard arguments, this algorithm requires the strong
    convexity constant mu_g of g.

    Parameters
    ----------
    x : primal variable
        This variable is both input and output of the method.
    f : functions
        Functionals Y[i] -> IR_infty that all have a convex conjugate with a
        proximal operator, i.e.
        f[i].convex_conj.proximal(sigma[i]) : Y[i] -> Y[i].
    g : function
        Functional X -> IR_infty that has a proximal operator, i.e.
        g.proximal(tau) : X -> X.
    A : functions
        Operators A[i] : X -> Y[i] that possess adjoints: A[i].adjoint
    tau : scalar
        Step size for primal variable.
    sigma : scalar
        Step size for dual variable.
    niter : int
        Number of iterations
    mu_g : scalar
        Strong convexity constant of g.

    Other Parameters
    ----------------
    y : dual variable
        Dual variable is part of a product space. By default equals 0.
    z : variable
        Adjoint of dual variable, z = A^* y. By default equals 0 if y = 0.
    prob: list
        List of probabilities that an index i is selected each iteration. By
        default this is uniform serial sampling, p_i = 1/n.
    fun_select : function
        Function that selects blocks at every iteration IN -> {1,...,n}. By
        default this is serial sampling, fun_select(k) selects an index
        i \in {1,...,n} with probability p_i.
    callback : callable, optional
        Function called with the current iterate after each iteration.

    References
    ----------
    [CERS2017] A. Chambolle, M. J. Ehrhardt, P. Richtarik and C.-B. Schoenlieb,
    *Stochastic Primal-Dual Hybrid Gradient Algorithm with Arbitrary Sampling
    and Imaging Applications*. ArXiv: http://arxiv.org/abs/1706.04957 (2017).
    """
    # Probabilities
    prob = kwargs.pop('prob', None)
    if prob is None:
        prob = [1 / len(A)] * len(A)

    # Selection function
    fun_select = kwargs.pop('fun_select', None)
    if fun_select is None:
        def fun_select(x):
            return [int(np.random.choice(len(A), 1, p=prob))]

    # Dual variable
    y = kwargs.pop('y', None)

    extra = [1 / p for p in prob]

    spdhg_generic(x, f, g, A, tau, sigma, niter, fun_select=fun_select,
                  extra=extra, mu_g=mu_g, y=y, **kwargs)


def spdhg_generic(x, f, g, A, tau, sigma, niter, **kwargs):
    r"""Computes a saddle point with a stochastic PDHG.

    This means, a solution (x*, y*), y* = (y*_1, ..., y*_n) such that

    (x*, y*) in arg min_x max_y sum_i=1^n <y_i, A_i> - f*[i](y_i) + g(x)

    where g : X -> IR_infty and f[i] : Y[i] -> IR_infty are convex, l.s.c. and
    proper functionals. For this algorithm, they all may be non-smooth and no
    strong convexity is assumed.

    Parameters
    ----------
    x : primal variable
        This variable is both input and output of the method.
    f : functions
        Functionals Y[i] -> IR_infty that all have a convex conjugate with a
        proximal operator, i.e.
        f[i].convex_conj.proximal(sigma[i]) : Y[i] -> Y[i].
    g : function
        Functional X -> IR_infty that has a proximal operator, i.e.
        g.proximal(tau) : X -> X.
    A : functions
        Operators A[i] : X -> Y[i] that possess adjoints: A[i].adjoint
    tau : scalar / vector / matrix
        Step size for primal variable. Note that the proximal operator of g
        has to be well-defined for this input.
    sigma : scalar
        Scalar / vector / matrix used as step size for dual variable. Note that
        the proximal operator related to f (see above) has to be well-defined
        for this input.
    niter : int
        Number of iterations

    Other Parameters
    ----------------
    y : dual variable, optional
        Dual variable is part of a product space. By default equals 0.
    z : variable, optional
        Adjoint of dual variable, z = A^* y. By default equals 0 if y = 0.
    mu_g : scalar
        Strong convexity constant of g.
    theta : scalar
        Global extrapolation factor.
    extra: list
        List of local extrapolation paramters for every index i. By default
        extra_i = 1.
    fun_select : function
        Function that selects blocks at every iteration IN -> {1,...,n}. By
        default this is serial uniform sampling, fun_select(k) selects an index
        i \in {1,...,n} with probability 1/n.
    callback : callable, optional
        Function called with the current iterate after each iteration.

    References
    ----------
    [CERS2017] A. Chambolle, M. J. Ehrhardt, P. Richtarik and C.-B. Schoenlieb,
    *Stochastic Primal-Dual Hybrid Gradient Algorithm with Arbitrary Sampling
    and Imaging Applications*. ArXiv: http://arxiv.org/abs/1706.04957 (2017).

    [E+2017] M. J. Ehrhardt, P. J. Markiewicz, P. Richtarik, J. Schott,
    A. Chambolle and C.-B. Schoenlieb, *Faster PET reconstruction with a
    stochastic primal-dual hybrid gradient method*. Wavelets and Sparsity XVII,
    58 (2017) http://doi.org/10.1117/12.2272946.
    """
    # Callback object
    callback = kwargs.pop('callback', None)
    if callback is not None and not callable(callback):
        raise TypeError('`callback` {} is not callable'
                        ''.format(callback))

    # Dual variable
    y = kwargs.pop('y', None)
    if y is None:
        y = A.range.zero()

    # Adjoint of dual variable
    z = kwargs.pop('z', None)
    if z is None:
        if y.norm() == 0:
            z = A.domain.zero()
        else:
            z = A.adjoint(y)

    # Strong convexity of g
    mu_g = kwargs.pop('mu_g', None)
    if mu_g is None:
        update_proximal_primal = False
    else:
        update_proximal_primal = True

    # Global extrapolation factor theta
    theta = kwargs.pop('theta', 1)

    # Second extrapolation factor
    extra = kwargs.pop('extra', None)
    if extra is None:
        extra = [1] * len(sigma)

    # Selection function
    fun_select = kwargs.pop('fun_select', None)
    if fun_select is None:
        def fun_select(x):
            return [int(np.random.choice(len(A), 1, p=1 / len(A)))]

    # Initialize variables
    z_relax = z.copy()
    dz = A.domain.element()
    y_old = A.range.element()

    # Save proximal operators
    proximal_dual_sigma = [fi.convex_conj.proximal(si)
                           for fi, si in zip(f, sigma)]
    proximal_primal_tau = g.proximal(tau)

    # run the iterations
    for k in range(niter):

        # select block
        selected = fun_select(k)

        # update primal variable
        # tmp = x - tau * z_relax; z_relax used as tmp variable
        z_relax.lincomb(1, x, -tau, z_relax)
        # x = prox(tmp)
        proximal_primal_tau(z_relax, out=x)

        # update extrapolation parameter theta
        if update_proximal_primal:
            theta = float(1 / np.sqrt(1 + 2 * mu_g * tau))

        # update dual variable and z, z_relax
        z_relax.assign(z)
        for i in selected:

            # save old yi
            y_old[i].assign(y[i])

            # tmp = Ai(x)
            A[i](x, out=y[i])

            # tmp = y_old + sigma_i * Ai(x)
            y[i].lincomb(1, y_old[i], sigma[i], y[i])

            # y[i]= prox(tmp)
            proximal_dual_sigma[i](y[i], out=y[i])

            # update adjoint of dual variable
            y_old[i].lincomb(-1, y_old[i], 1, y[i])
            A[i].adjoint(y_old[i], out=dz)
            z += dz

            # compute extrapolation
            z_relax.lincomb(1, z_relax, 1 + theta * extra[i], dz)

        # update the step sizes tau and sigma for acceleration
        if update_proximal_primal:
            for i in range(len(sigma)):
                sigma[i] /= theta
            tau *= theta

            proximal_dual_sigma = [fi.convex_conj.proximal(si)
                                   for fi, si in zip(f, sigma)]
            proximal_primal_tau = g.proximal(tau)

        if callback is not None:
            callback([x, y])


def da_spdhg(x, f, g, A, tau, sigma_tilde, niter, mu, **kwargs):
    r"""Computes a saddle point with a PDHG and dual acceleration.

    It therefore requires the functionals f*_i to be mu[i] strongly convex.

    Parameters
    ----------
    x : primal variable
        This variable is both input and output of the method.
    f : functions
        Functionals Y[i] -> IR_infty that all have a convex conjugate with a
        proximal operator, i.e.
        f[i].convex_conj.proximal(sigma[i]) : Y[i] -> Y[i].
    g : function
        Functional X -> IR_infty that has a proximal operator, i.e.
        g.proximal(tau) : X -> X.
    A : functions
        Operators A[i] : X -> Y[i] that possess adjoints: A[i].adjoint
    tau : scalar
        Initial step size for primal variable.
    sigma_tilde : scalar
        Related to initial step size for dual variable.
    niter : int
        Number of iterations
    mu: list
        List of strong convexity constants of f*, i.e. mu[i] is the strong
        convexity constant of f*[i].

    Other Parameters
    ----------------
    y: dual variable
        Dual variable is part of a product space
    z: variable
        Adjoint of dual variable, z = A^* y.
    prob: list
        List of probabilities that an index i is selected each iteration. By
        default this is uniform serial sampling, p_i = 1/n.
    fun_select : function
        Function that selects blocks at every iteration IN -> {1,...,n}. By
        default this is serial sampling, fun_select(k) selects an index
        i \in {1,...,n} with probability p_i.
    extra: list
        List of local extrapolation paramters for every index i. By default
        extra_i = 1 / p_i.
    callback : callable, optional
        Function called with the current iterate after each iteration.

    References
    ----------
    [CERS2017] A. Chambolle, M. J. Ehrhardt, P. Richtarik and C.-B. Schoenlieb,
    *Stochastic Primal-Dual Hybrid Gradient Algorithm with Arbitrary Sampling
    and Imaging Applications*. ArXiv: http://arxiv.org/abs/1706.04957 (2017).
    """
    # Callback object
    callback = kwargs.pop('callback', None)
    if callback is not None and not callable(callback):
        raise TypeError('`callback` {} is not callable'
                        ''.format(callback))

    # Probabilities
    prob = kwargs.pop('prob', None)
    if prob is None:
        prob = [1 / len(A)] * len(A)

    # Selection function
    fun_select = kwargs.pop('fun_select', None)
    if fun_select is None:
        def fun_select(x):
            return [int(np.random.choice(len(A), 1, p=prob))]

    # Dual variable
    y = kwargs.pop('y', None)
    if y is None:
        y = A.range.zero()

    # Adjoint of dual variable
    z = kwargs.pop('z', None)
    if z is None and y.norm() == 0:
        z = A.domain.zero()

    # Extrapolation
    extra = kwargs.pop('extra', None)
    if extra is None:
        extra = [1 / p for p in prob]

    # Initialize variables
    z_relax = z.copy()
    dz = A.domain.element()
    y_old = A.range.element()

    # Save proximal operators
    prox_dual = [fi.convex_conj.proximal for fi in f]
    prox_primal = g.proximal

    # run the iterations
    for k in range(niter):

        # select block
        selected = fun_select(k)

        # update extrapolation parameter theta
        theta = float(1 / np.sqrt(1 + 2 * sigma_tilde))

        # update primal variable
        # tmp = x - tau * z_relax; z_relax used as tmp variable
        z_relax.lincomb(1, x, -tau, z_relax)
        # x = prox(tmp)
        prox_primal(tau)(z_relax, out=x)

        # update dual variable and z, z_relax
        z_relax.assign(z)
        for i in selected:

            # compute the step sizes sigma_i based on sigma_tilde
            sigma_i = sigma_tilde / (
                mu[i] * (prob[i] - 2 * (1 - prob[i]) * sigma_tilde))

            # save old yi
            y_old[i].assign(y[i])

            # tmp = Ai(x)
            A[i](x, out=y[i])

            # tmp = y_old + sigma_i * Ai(x)
            y[i].lincomb(1, y_old[i], sigma_i, y[i])

            # yi++ = fi*.prox_sigmai(yi)
            prox_dual[i](sigma_i)(y[i], out=y[i])

            # update adjoint of dual variable
            y_old[i].lincomb(-1, y_old[i], 1, y[i])
            A[i].adjoint(y_old[i], out=dz)
            z += dz

            # compute extrapolation
            z_relax.lincomb(1, z_relax, 1 + theta * extra[i], dz)

        # update the step sizes tau and sigma_tilde for acceleration
        sigma_tilde *= theta
        tau /= theta

        if callback is not None:
            callback([x, y])


def spdhg_pesquet(x, f, g, A, tau, sigma, niter, **kwargs):
    r"""Computes a saddle point with a stochstic variant of PDHG [PR2015].

    Parameters
    ----------
    x : primal variable
        This variable is both input and output of the method.
    f : functions
        Functionals Y[i] -> IR_infty that all have a convex conjugate with a
        proximal operator, i.e.
        f[i].convex_conj.proximal(sigma[i]) : Y[i] -> Y[i].
    g : function
        Functional X -> IR_infty that has a proximal operator, i.e.
        g.proximal(tau) : X -> X.
    A : functions
        Operators A[i] : X -> Y[i] that possess adjoints: A[i].adjoint
    tau : scalar / vector / matrix
        Step size for primal variable. Note that the proximal operator of g
        has to be well-defined for this input.
    sigma : list
        List of scalars / vectors / matrices used as step sizes for the dual
        variable. Note that the proximal operators related to f (see above)
        have to be well-defined for this input.
    niter : int
        Number of iterations

    Other Parameters
    ----------------
    y: dual variable
        Dual variable is part of a product space
    z: variable
        Adjoint of dual variable, z = A^* y.
    fun_select : function
        Function that selects blocks at every iteration IN -> {1,...,n}. By
        default this is uniform serial sampling, fun_select(k) selects
        uniformly an i \in {1,...,n}.
    callback : callable, optional
        Function called with the current iterate after each iteration.

    References
    ----------
    [PR2015] J.-C. Pesquet and A. Repetti. *A Class of Randomized Primal-Dual
    Algorithms for Distributed Optimization*.
    ArXiv: http://arxiv.org/abs/1406.6404 (2015).
    """
    # Callback object
    callback = kwargs.pop('callback', None)
    if callback is not None and not callable(callback):
        raise TypeError('`callback` {} is not callable'
                        ''.format(callback))

    # Selection function
    fun_select = kwargs.pop('fun_select', None)
    if fun_select is None:
        def fun_select(x):
            return [int(np.random.choice(len(A), 1, p=1 / len(A)))]

    # Dual variable
    y = kwargs.pop('y', None)
    if y is None:
        y = A.range.zero()

    # Adjoint of dual variable
    z = kwargs.pop('z', None)
    if z is None and y.norm() == 0:
        z = A.domain.zero()

    # Save proximal operators
    prox_f = [f[i].convex_conj.proximal(sigma[i]) for i in range(len(sigma))]
    prox_g = g.proximal(tau)

    # Initialize variables
    x_relax = x.copy()
    y_old = A.range.element()
    dz = z.copy()

    # run the iterations
    for k in range(niter):
        x_relax.lincomb(-1, x)

        # update primal variable
        x.lincomb(1, x, -tau, z)
        prox_g(x, out=x)

        # compute extrapolation
        x_relax.lincomb(1, x_relax, 2, x)

        # select block
        selected = fun_select(k)

        # update dual variable and adj_y
        for i in selected:
            # update dual variable
            y_old[i].assign(y[i])

            A[i](x_relax, out=y[i])
            y[i] *= sigma[i]
            y[i] += y_old[i]
            prox_f[i](y[i], out=y[i])

            # update adjoint of dual variable
            A[i].adjoint(y[i] - y_old[i], out=dz)
            z += dz

        if callback is not None:
            callback([x, y])



# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Utilities mainly for internal use."""

# Python imports
from itertools import zip_longest
from contextlib import contextmanager
# ODL import
from odl.core.array_API_support.array_creation import asarray
from odl.core.array_API_support.utils import get_array_and_backend
from odl.core.util.dtype_utils import _universal_dtype_identifier
# Third-party import
import numpy as np

__all__ = (   
    'is_string',
    'dtype_repr',
    'dtype_str',
    'REPR_PRECISION',
    'indent',
    'dedent',
    'npy_printoptions',
    'array_str',
    'signature_string',
    'signature_string_parts',
    'repr_string',
    'attribute_repr_string',
    'method_repr_string',
)

def is_string(obj):
    """Return ``True`` if ``obj`` behaves like a string, ``False`` else."""
    return isinstance(obj, str)

def dtype_repr(dtype):
    return f"'{dtype_str(dtype)}'"

def dtype_str(dtype):
    return f"{_universal_dtype_identifier(dtype)}"

REPR_PRECISION = 4

def indent(string, indent_str='    '):
    """Return a copy of ``string`` indented by ``indent_str``.

    Parameters
    ----------
    string : str
        Text that should be indented.
    indent_str : str, optional
        String to be inserted before each new line. The default is to
        indent by 4 spaces.

    Returns
    -------
    indented : str
        The indented text.

    Examples
    --------
    >>> text = '''This is line 1.
    ... Next line.
    ... And another one.'''
    >>> print(text)
    This is line 1.
    Next line.
    And another one.
    >>> print(indent(text))
        This is line 1.
        Next line.
        And another one.

    Indenting by random stuff:

    >>> print(indent(text, indent_str='<->'))
    <->This is line 1.
    <->Next line.
    <->And another one.
    """
    return '\n'.join(indent_str + row for row in string.splitlines())


def dedent(string, indent_str='   ', max_levels=None):
    """Revert the effect of indentation.

    Examples
    --------
    Remove a simple one-level indentation:

    >>> text = '''<->This is line 1.
    ... <->Next line.
    ... <->And another one.'''
    >>> print(text)
    <->This is line 1.
    <->Next line.
    <->And another one.
    >>> print(dedent(text, '<->'))
    This is line 1.
    Next line.
    And another one.

    Multiple levels of indentation:

    >>> text = '''<->Level 1.
    ... <-><->Level 2.
    ... <-><-><->Level 3.'''
    >>> print(text)
    <->Level 1.
    <-><->Level 2.
    <-><-><->Level 3.
    >>> print(dedent(text, '<->'))
    Level 1.
    <->Level 2.
    <-><->Level 3.

    >>> text = '''<-><->Level 2.
    ... <-><-><->Level 3.'''
    >>> print(text)
    <-><->Level 2.
    <-><-><->Level 3.
    >>> print(dedent(text, '<->'))
    Level 2.
    <->Level 3.
    >>> print(dedent(text, '<->', max_levels=1))
    <->Level 2.
    <-><->Level 3.
    """
    if len(indent_str) == 0:
        return string

    lines = string.splitlines()

    # Determine common (minimum) number of indentation levels, capped at
    # `max_levels` if given
    def num_indents(line):
        max_num = int(np.ceil(len(line) / len(indent_str)))

        i = 0  # set for the case the loop is not run (`max_num == 0`)
        for i in range(max_num):
            if line.startswith(indent_str):
                line = line[len(indent_str):]
            else:
                break

        return i

    num_levels = num_indents(min(lines, key=num_indents))
    if max_levels is not None:
        num_levels = min(num_levels, max_levels)

    # Dedent
    dedent_len = num_levels * len(indent_str)
    return '\n'.join(line[dedent_len:] for line in lines)

@contextmanager
def npy_printoptions(**extra_opts):
    """Context manager to temporarily set NumPy print options.

    See Also
    --------
    numpy.get_printoptions
    numpy.set_printoptions

    Examples
    --------
    >>> print(np.array([np.nan, 1.00001]))
    [     nan  1.00001]
    >>> with npy_printoptions(precision=3):
    ...     print(np.array([np.nan, 1.00001]))
    [ nan   1.]
    >>> with npy_printoptions(nanstr='whoah!'):
    ...     print(np.array([np.nan, 1.00001]))
    [  whoah!  1.00001]
    """
    orig_opts = np.get_printoptions()

    try:
        new_opts = orig_opts.copy()
        new_opts.update(extra_opts)
        np.set_printoptions(**new_opts)
        yield

    finally:
        np.set_printoptions(**orig_opts)


def array_str(a, nprint=6):
    """Stringification of an array.

    Parameters
    ----------
    a : `array-like`
        The array to print.
    nprint : int, optional
        Maximum number of elements to print per axis in ``a``. For larger
        arrays, a summary is printed, with ``nprint // 2`` elements on
        each side and ``...`` in the middle (per axis).

    Examples
    --------
    Printing 1D arrays:

    >>> print(array_str(np.arange(4)))
    [0, 1, 2, 3]
    >>> print(array_str(np.arange(10)))
    [0, 1, 2, ..., 7, 8, 9]
    >>> print(array_str(np.arange(10), nprint=10))
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    For 2D and higher, the ``nprint`` limitation applies per axis:

    >>> print(array_str(np.arange(24).reshape(4, 6)))
    [[ 0,  1,  2,  3,  4,  5],
     [ 6,  7,  8,  9, 10, 11],
     [12, 13, 14, 15, 16, 17],
     [18, 19, 20, 21, 22, 23]]
    >>> print(array_str(np.arange(32).reshape(4, 8)))
    [[ 0,  1,  2, ...,  5,  6,  7],
     [ 8,  9, 10, ..., 13, 14, 15],
     [16, 17, 18, ..., 21, 22, 23],
     [24, 25, 26, ..., 29, 30, 31]]
    >>> print(array_str(np.arange(32).reshape(8, 4)))
    [[ 0,  1,  2,  3],
     [ 4,  5,  6,  7],
     [ 8,  9, 10, 11],
     ...,
     [20, 21, 22, 23],
     [24, 25, 26, 27],
     [28, 29, 30, 31]]
    >>> print(array_str(np.arange(64).reshape(8, 8)))
    [[ 0,  1,  2, ...,  5,  6,  7],
     [ 8,  9, 10, ..., 13, 14, 15],
     [16, 17, 18, ..., 21, 22, 23],
     ...,
     [40, 41, 42, ..., 45, 46, 47],
     [48, 49, 50, ..., 53, 54, 55],
     [56, 57, 58, ..., 61, 62, 63]]

    Printing of empty arrays and 0D arrays:

    >>> print(array_str(np.array([])))  # 1D, size=0
    []
    >>> print(array_str(np.array(1.0)))  # 0D, size=1
    1.0

    Small deviations from round numbers will be suppressed:

    >>> # 2.0000000000000004 in double precision
    >>> print(array_str((np.array([2.0]) ** 0.5) ** 2))
    [ 2.]
    """
    a = asarray(a)
    a, backend = get_array_and_backend(a)
    a = backend.to_numpy(a)
    max_shape = tuple(n if n < nprint else nprint for n in a.shape)
    with npy_printoptions(threshold=int(np.prod(max_shape)),
                          edgeitems=nprint // 2,
                          suppress=True):
        a_str = np.array2string(a, separator=', ')
    return a_str


def signature_string(posargs, optargs, sep=', ', mod='!r'):
    """Return a stringified signature from given arguments.

    Parameters
    ----------
    posargs : sequence
        Positional argument values, always included in the returned string.
        They appear in the string as (roughly)::

            sep.join(str(arg) for arg in posargs)

    optargs : sequence of 3-tuples
        Optional arguments with names and defaults, given in the form::

            [(name1, value1, default1), (name2, value2, default2), ...]

        Only those parameters that are different from the given default
        are included as ``name=value`` keyword pairs.

        **Note:** The comparison is done by using ``if value == default:``,
        which is not valid for, e.g., NumPy arrays.

    sep : string or sequence of strings, optional
        Separator(s) for the argument strings. A provided single string is
        used for all joining operations.
        A given sequence must have 3 entries ``pos_sep, opt_sep, part_sep``.
        The ``pos_sep`` and ``opt_sep`` strings are used for joining the
        respective sequences of argument strings, and ``part_sep`` joins
        these two joined strings.
    mod : string or callable or sequence, optional
        Format modifier(s) for the argument strings.
        In its most general form, ``mod`` is a sequence of 2 sequences
        ``pos_mod, opt_mod`` with ``len(pos_mod) == len(posargs)`` and
        ``len(opt_mod) == len(optargs)``. Each entry ``m`` in those sequences
        can be eiter a string, resulting in the following stringification
        of ``arg``::

            arg_fmt = {{{}}}.format(m)
            arg_str = arg_fmt.format(arg)

        For a callable ``to_str``, the stringification is simply
        ``arg_str = to_str(arg)``.

        The entries ``pos_mod, opt_mod`` of ``mod`` can also be strings
        or callables instead of sequences, in which case the modifier
        applies to all corresponding arguments.

        Finally, if ``mod`` is a string or callable, it is applied to
        all arguments.

        The default behavior is to apply the "{!r}" (``repr``) conversion.
        For floating point scalars, the number of digits printed is
        determined by the ``precision`` value in NumPy's printing options,
        which can be temporarily modified with `npy_printoptions`.

    Returns
    -------
    signature : string
        Stringification of a signature, typically used in the form::

            '{}({})'.format(self.__class__.__name__, signature)

    Examples
    --------
    Usage with non-trivial entries in both sequences, with a typical
    use case:

    >>> posargs = [1, 'hello', None]
    >>> optargs = [('dtype', 'float32', 'float64')]
    >>> signature_string(posargs, optargs)
    "1, 'hello', None, dtype='float32'"
    >>> '{}({})'.format('MyClass', signature_string(posargs, optargs))
    "MyClass(1, 'hello', None, dtype='float32')"

    Empty sequences and optargs values equal to default are omitted:

    >>> posargs = ['hello']
    >>> optargs = [('size', 1, 1)]
    >>> signature_string(posargs, optargs)
    "'hello'"
    >>> posargs = []
    >>> optargs = [('size', 2, 1)]
    >>> signature_string(posargs, optargs)
    'size=2'
    >>> posargs = []
    >>> optargs = [('size', 1, 1)]
    >>> signature_string(posargs, optargs)
    ''

    Using a different separator, globally or per argument "category":

    >>> posargs = [1, 'hello', None]
    >>> optargs = [('dtype', 'float32', 'float64'),
    ...            ('order', 'F', 'C')]
    >>> signature_string(posargs, optargs)
    "1, 'hello', None, dtype='float32', order='F'"
    >>> signature_string(posargs, optargs, sep=(',', ',', ', '))
    "1,'hello',None, dtype='float32',order='F'"

    Using format modifiers:

    >>> posargs = ['hello', 2.345]
    >>> optargs = [('extent', 1.442, 1.0), ('spacing', 0.0151, 1.0)]
    >>> signature_string(posargs, optargs)
    "'hello', 2.345, extent=1.442, spacing=0.0151"
    >>> # Print only two significant digits for all arguments.
    >>> # NOTE: this also affects the string!
    >>> mod = ':.2'
    >>> signature_string(posargs, optargs, mod=mod)
    'he, 2.3, extent=1.4, spacing=0.015'
    >>> mod = [['', ''], [':.3', ':.2']]  # one modifier per argument
    >>> signature_string(posargs, optargs, mod=mod)
    "'hello', 2.345, extent=1.44, spacing=0.015"

    Using callables for stringification:

    >>> posargs = ['arg1', np.ones(3)]
    >>> optargs = []
    >>> signature_string(posargs, optargs, mod=[['', array_str], []])
    "'arg1', [ 1., 1., 1.]"

    The number of printed digits in floating point numbers can be changed
    with `npy_printoptions`:

    >>> posargs = ['hello', 0.123456789012345]
    >>> optargs = [('extent', 1.234567890123456, 1.0)]
    >>> signature_string(posargs, optargs)  # default is 8 digits
    "'hello', 0.12345679, extent=1.2345679"
    >>> with npy_printoptions(precision=2):
    ...     sig_str = signature_string(posargs, optargs)
    >>> sig_str
    "'hello', 0.12, extent=1.2"
    """
    # Define the separators for the two possible cases
    if is_string(sep):
        pos_sep = opt_sep = part_sep = sep
    else:
        pos_sep, opt_sep, part_sep = sep

    # Get the stringified parts
    posargs_conv, optargs_conv = signature_string_parts(posargs, optargs, mod)

    # Join the arguments using the separators
    parts = []
    if posargs_conv:
        parts.append(pos_sep.join(argstr for argstr in posargs_conv))
    if optargs_conv:
        parts.append(opt_sep.join(optargs_conv))

    return part_sep.join(parts)


def signature_string_parts(posargs, optargs, mod='!r'):
    """Return stringified arguments as tuples.

    Parameters
    ----------
    posargs : sequence
        Positional argument values, always included in the returned string
        tuple.
    optargs : sequence of 3-tuples
        Optional arguments with names and defaults, given in the form::

            [(name1, value1, default1), (name2, value2, default2), ...]

        Only those parameters that are different from the given default
        are included as ``name=value`` keyword pairs.

        **Note:** The comparison is done by using ``if value == default:``,
        which is not valid for, e.g., NumPy arrays.

    mod : string or callable or sequence, optional
        Format modifier(s) for the argument strings.
        In its most general form, ``mod`` is a sequence of 2 sequences
        ``pos_mod, opt_mod`` with ``len(pos_mod) == len(posargs)`` and
        ``len(opt_mod) == len(optargs)``. Each entry ``m`` in those sequences
        can be a string, resulting in the following stringification
        of ``arg``::

            arg_fmt = {{{}}}.format(m)
            arg_str = arg_fmt.format(arg)

        For a callable ``to_str``, the stringification is simply
        ``arg_str = to_str(arg)``.

        The entries ``pos_mod, opt_mod`` of ``mod`` can also be strings
        or callables instead of sequences, in which case the modifier
        applies to all corresponding arguments.

        Finally, if ``mod`` is a string or callable, it is applied to
        all arguments.

        The default behavior is to apply the "{!r}" (``repr``) conversion.
        For floating point scalars, the number of digits printed is
        determined by the ``precision`` value in NumPy's printing options,
        which can be temporarily modified with `npy_printoptions`.

    Returns
    -------
    pos_strings : tuple of str
        The stringified positional arguments.
    opt_strings : tuple of str
        The stringified optional arguments, not including the ones
        equal to their respective defaults.
    """
    # Convert modifiers to 2-sequence of sequence of strings
    if is_string(mod) or callable(mod):
        pos_mod = opt_mod = mod
    else:
        pos_mod, opt_mod = mod

    mods = []
    for m, args in zip((pos_mod, opt_mod), (posargs, optargs)):
        if is_string(m) or callable(m):
            mods.append([m] * len(args))
        else:
            if len(m) == 1:
                mods.append(m * len(args))
            elif len(m) == len(args):
                mods.append(m)
            else:
                raise ValueError('sequence length mismatch: '
                                 'len({}) != len({})'.format(m, args))

    pos_mod, opt_mod = mods
    precision = np.get_printoptions()['precision']

    # Stringify values, treating strings specially
    posargs_conv = []
    for arg, modifier in zip(posargs, pos_mod):
        if callable(modifier):
            posargs_conv.append(modifier(arg))
        elif is_string(arg):
            # Preserve single quotes for strings by default
            if modifier:
                fmt = '{{{}}}'.format(modifier)
            else:
                fmt = "'{}'"
            posargs_conv.append(fmt.format(arg))
        elif np.isscalar(arg) and str(arg) in ('inf', 'nan'):
            # Make sure the string quotes are added
            posargs_conv.append("'{}'".format(arg))
        elif (np.isscalar(arg) and
              np.array(arg).real.astype('int64') != arg and
              modifier in ('', '!s', '!r')):
            # Floating point value, use numpy print option 'precision'
            fmt = '{{:.{}}}'.format(precision)
            posargs_conv.append(fmt.format(arg))
        else:
            # All non-string types are passed through a format conversion
            fmt = '{{{}}}'.format(modifier)
            posargs_conv.append(fmt.format(arg))

    # Build 'key=value' strings for values that are not equal to default
    optargs_conv = []
    for (name, value, default), modifier in zip(optargs, opt_mod):
        if value == default:
            # Don't include
            continue

        # See above on str and repr
        if callable(modifier):
            optargs_conv.append('{}={}'.format(name, modifier(value)))
        elif is_string(value):
            if modifier:
                fmt = '{{{}}}'.format(modifier)
            else:
                fmt = "'{}'"
            value_str = fmt.format(value)
            optargs_conv.append('{}={}'.format(name, value_str))
        elif np.isscalar(value) and str(value) in ('inf', 'nan'):
            # Make sure the string quotes are added
            optargs_conv.append("{}='{}'".format(name, value))
        elif (np.isscalar(value) and
              np.array(value).real.astype('int64') != value and
              modifier in ('', '!s', '!r')):
            fmt = '{{:.{}}}'.format(precision)
            value_str = fmt.format(value)
            optargs_conv.append('{}={}'.format(name, value_str))
        else:
            fmt = '{{{}}}'.format(modifier)
            value_str = fmt.format(value)
            optargs_conv.append('{}={}'.format(name, value_str))

    return tuple(posargs_conv), tuple(optargs_conv)


def _separators(strings, linewidth):
    """Return separators that keep joined strings within the line width."""
    if len(strings) <= 1:
        return ()

    indent_len = 4
    separators = []
    cur_line_len = indent_len + len(strings[0]) + 1
    if cur_line_len + 2 <= linewidth and '\n' not in strings[0]:
        # Next string might fit on same line
        separators.append(', ')
        cur_line_len += 1  # for the extra space
    else:
        # Use linebreak if string contains newline or doesn't fit
        separators.append(',\n')
        cur_line_len = indent_len

    for i, s in enumerate(strings[1:-1]):
        cur_line_len += len(s) + 1

        if '\n' in s:
            # Use linebreak before and after if string contains newline
            separators[i] = ',\n'
            cur_line_len = indent_len
            separators.append(',\n')

        elif cur_line_len + 2 <= linewidth:
            # This string fits, next one might also fit on same line
            separators.append(', ')
            cur_line_len += 1  # for the extra space

        elif cur_line_len <= linewidth:
            # This string fits, but next one won't
            separators.append(',\n')
            cur_line_len = indent_len

        else:
            # This string doesn't fit but has no newlines in it
            separators[i] = ',\n'
            cur_line_len = indent_len + len(s) + 1

            # Need to determine again what should come next
            if cur_line_len + 2 <= linewidth:
                # Next string might fit on same line
                separators.append(', ')
            else:
                separators.append(',\n')

    cur_line_len += len(strings[-1])
    if cur_line_len + 1 > linewidth or '\n' in strings[-1]:
        # This string and a comma don't fit on this line
        separators[-1] = ',\n'

    return tuple(separators)


def repr_string(outer_string, inner_strings, allow_mixed_seps=True):
    r"""Return a pretty string for ``repr``.

    The returned string is formatted such that it does not extend
    beyond the line boundary if avoidable. The line width is taken from
    NumPy's printing options that can be retrieved with
    `numpy.get_printoptions`. They can be temporarily overridden
    using the `npy_printoptions` context manager. See Examples for details.

    Parameters
    ----------
    outer_string : str
        Name of the class or function that should be printed outside
        the parentheses.
    inner_strings : sequence of sequence of str
        Stringifications of the positional and optional arguments.
        This is usually the return value of `signature_string_parts`.
    allow_mixed_seps : bool, optional
        If ``False`` and the string does not fit on one line, use
        ``',\n'`` to separate all strings.
        By default, a mixture of ``', '`` and ``',\n'`` is used to fit
        as much on one line as possible.

        In case some of the ``inner_strings`` span multiple lines, it is
        usually advisable to set ``allow_mixed_seps`` to ``False`` since
        the result tends to be more readable that way.

    Returns
    -------
    repr_string : str
        Full string that can be returned by a class' ``__repr__`` method.

    Examples
    --------
    Things that fit into one line are printed on one line:

    >>> outer_string = 'MyClass'
    >>> inner_strings = [('1', "'hello'", 'None'),
    ...                  ("dtype='float32'",)]
    >>> print(repr_string(outer_string, inner_strings))
    MyClass(1, 'hello', None, dtype='float32')

    Otherwise, if a part of ``inner_strings`` fits on a line of its own,
    it is printed on one line, but separated from the other part with
    a line break:

    >>> outer_string = 'MyClass'
    >>> inner_strings = [('2.0', "'this_is_a_very_long_argument_string'"),
    ...                  ("long_opt_arg='another_quite_long_string'",)]
    >>> print(repr_string(outer_string, inner_strings))
    MyClass(
        2.0, 'this_is_a_very_long_argument_string',
        long_opt_arg='another_quite_long_string'
    )

    If those parts are themselves too long, they are broken down into
    several lines:

    >>> outer_string = 'MyClass'
    >>> inner_strings = [("'this_is_a_very_long_argument_string'",
    ...                   "'another_very_long_argument_string'"),
    ...                  ("long_opt_arg='another_quite_long_string'",
    ...                   "long_opt2_arg='this_wont_fit_on_one_line_either'")]
    >>> print(repr_string(outer_string, inner_strings))
    MyClass(
        'this_is_a_very_long_argument_string',
        'another_very_long_argument_string',
        long_opt_arg='another_quite_long_string',
        long_opt2_arg='this_wont_fit_on_one_line_either'
    )

    The usage of mixed separators to optimally use horizontal space can
    be disabled by setting ``allow_mixed_seps=False``:

    >>> outer_string = 'MyClass'
    >>> inner_strings = [('2.0', "'this_is_a_very_long_argument_string'"),
    ...                  ("long_opt_arg='another_quite_long_string'",)]
    >>> print(repr_string(outer_string, inner_strings, allow_mixed_seps=False))
    MyClass(
        2.0,
        'this_is_a_very_long_argument_string',
        long_opt_arg='another_quite_long_string'
    )

    With the ``npy_printoptions`` context manager, the available line
    width can be changed:

    >>> outer_string = 'MyClass'
    >>> inner_strings = [('1', "'hello'", 'None'),
    ...                  ("dtype='float32'",)]
    >>> with npy_printoptions(linewidth=20):
    ...     print(repr_string(outer_string, inner_strings))
    MyClass(
        1, 'hello',
        None,
        dtype='float32'
    )
    """
    linewidth = np.get_printoptions()['linewidth']
    pos_strings, opt_strings = inner_strings
    # Length of the positional and optional argument parts of the signature,
    # including separators `', '`
    pos_sig_len = (sum(len(pstr) for pstr in pos_strings) +
                   2 * max((len(pos_strings) - 1), 0))
    opt_sig_len = (sum(len(pstr) for pstr in opt_strings) +
                   2 * max((len(opt_strings) - 1), 0))

    # Length of the one-line string, including 2 for the parentheses and
    # 2 for the joining ', '
    repr_len = len(outer_string) + 2 + pos_sig_len + 2 + opt_sig_len

    if repr_len <= linewidth and not any('\n' in s
                                         for s in pos_strings + opt_strings):
        # Everything fits on one line
        fmt = '{}({})'
        pos_str = ', '.join(pos_strings)
        opt_str = ', '.join(opt_strings)
        parts_sep = ', '
    else:
        # Need to split lines in some way
        fmt = '{}(\n{}\n)'

        if not allow_mixed_seps:
            pos_separators = [',\n'] * (len(pos_strings) - 1)
        else:
            pos_separators = _separators(pos_strings, linewidth)
        if len(pos_strings) == 0:
            pos_str = ''
        else:
            pos_str = pos_strings[0]
            for s, sep in zip(pos_strings[1:], pos_separators):
                pos_str = sep.join([pos_str, s])

        if not allow_mixed_seps:
            opt_separators = [',\n'] * (len(opt_strings) - 1)
        else:
            opt_separators = _separators(opt_strings, linewidth)
        if len(opt_strings) == 0:
            opt_str = ''
        else:
            opt_str = opt_strings[0]
            for s, sep in zip(opt_strings[1:], opt_separators):
                opt_str = sep.join([opt_str, s])

        # Check if we can put both parts on one line. This requires their
        # concatenation including 4 for indentation and 2 for ', ' to
        # be less than the line width. And they should contain no newline.
        if pos_str and opt_str:
            inner_len = 4 + len(pos_str) + 2 + len(opt_str)
        elif (pos_str and not opt_str) or (opt_str and not pos_str):
            inner_len = 4 + len(pos_str) + len(opt_str)
        else:
            inner_len = 0

        if (not allow_mixed_seps or
                any('\n' in s for s in [pos_str, opt_str]) or
                inner_len > linewidth):
            parts_sep = ',\n'
            pos_str = indent(pos_str)
            opt_str = indent(opt_str)
        else:
            parts_sep = ', '
            pos_str = indent(pos_str)
            # Don't indent `opt_str`

    parts = [s for s in [pos_str, opt_str] if s.strip()]  # ignore empty
    inner_string = parts_sep.join(parts)
    return fmt.format(outer_string, inner_string)


def attribute_repr_string(inst_str, attr_str):
    """Return a repr string for an attribute that respects line width.

    Parameters
    ----------
    inst_str : str
        Stringification of a class instance.
    attr_str : str
        Name of the attribute (not including the ``'.'``).

    Returns
    -------
    attr_repr_str : str
        Concatenation of the two strings in a way that the line width
        is respected.

    Examples
    --------
    >>> inst_str = 'rn((2, 3))'
    >>> attr_str = 'byaxis'
    >>> print(attribute_repr_string(inst_str, attr_str))
    rn((2, 3)).byaxis
    >>> inst_str = 'MyClass()'
    >>> attr_str = 'attr_name'
    >>> print(attribute_repr_string(inst_str, attr_str))
    MyClass().attr_name
    >>> inst_str = 'MyClass'
    >>> attr_str = 'class_attr'
    >>> print(attribute_repr_string(inst_str, attr_str))
    MyClass.class_attr
    >>> long_inst_str = (
    ...     "MyClass('long string that will definitely trigger a line break')"
    ... )
    >>> long_attr_str = 'long_attribute_name'
    >>> print(attribute_repr_string(long_inst_str, long_attr_str))
    MyClass(
        'long string that will definitely trigger a line break'
    ).long_attribute_name
    """
    linewidth = np.get_printoptions()['linewidth']
    if (len(inst_str) + 1 + len(attr_str) <= linewidth or
            '(' not in inst_str):
        # Instance string + dot + attribute string fit in one line or
        # no parentheses -> keep instance string as-is and append attr string
        parts = [inst_str, attr_str]
    else:
        # TODO(kohr-h): use `maxsplit=1` kwarg, not supported in Py 2
        left, rest = inst_str.split('(', 1)
        right, middle = rest[::-1].split(')', 1)
        middle, right = middle[::-1], right[::-1]

        if middle.startswith('\n') and middle.endswith('\n'):
            # Already on multiple lines
            new_inst_str = inst_str
        else:
            init_parts = [left]
            if middle:
                init_parts.append(indent(middle))
            new_inst_str = '(\n'.join(init_parts) + '\n)' + right
        parts = [new_inst_str, attr_str]

    return '.'.join(parts)


def method_repr_string(inst_str, meth_str, arg_strs=None,
                       allow_mixed_seps=True):
    r"""Return a repr string for a method that respects line width.

    This function is useful to generate a ``repr`` string for a derived
    class that is created through a method, for instance ::

        functional.translated(x)

    as a better way of representing ::

        FunctionalTranslation(functional, x)

    Parameters
    ----------
    inst_str : str
        Stringification of a class instance.
    meth_str : str
        Name of the method (not including the ``'.'``).
    arg_strs : sequence of str, optional
        Stringification of the arguments to the method.
    allow_mixed_seps : bool, optional
        If ``False`` and the argument strings do not fit on one line, use
        ``',\n'`` to separate all strings.
        By default, a mixture of ``', '`` and ``',\n'`` is used to fit
        as much on one line as possible.

        In case some of the ``arg_strs`` span multiple lines, it is
        usually advisable to set ``allow_mixed_seps`` to ``False`` since
        the result tends to be more readable that way.

    Returns
    -------
    meth_repr_str : str
        Concatenation of all strings in a way that the line width
        is respected.

    Examples
    --------
    >>> inst_str = 'MyClass'
    >>> meth_str = 'empty'
    >>> arg_strs = []
    >>> print(method_repr_string(inst_str, meth_str, arg_strs))
    MyClass.empty()
    >>> inst_str = 'MyClass'
    >>> meth_str = 'fromfile'
    >>> arg_strs = ["'tmpfile.txt'"]
    >>> print(method_repr_string(inst_str, meth_str, arg_strs))
    MyClass.fromfile('tmpfile.txt')
    >>> inst_str = "MyClass('init string')"
    >>> meth_str = 'method'
    >>> arg_strs = ['2.0']
    >>> print(method_repr_string(inst_str, meth_str, arg_strs))
    MyClass('init string').method(2.0)
    >>> long_inst_str = (
    ...     "MyClass('long string that will definitely trigger a line break')"
    ... )
    >>> meth_str = 'method'
    >>> long_arg1 = "'long argument string that should come on the next line'"
    >>> arg2 = 'param1=1'
    >>> arg3 = 'param2=2.0'
    >>> arg_strs = [long_arg1, arg2, arg3]
    >>> print(method_repr_string(long_inst_str, meth_str, arg_strs))
    MyClass(
        'long string that will definitely trigger a line break'
    ).method(
        'long argument string that should come on the next line',
        param1=1, param2=2.0
    )
    >>> print(method_repr_string(long_inst_str, meth_str, arg_strs,
    ...                          allow_mixed_seps=False))
    MyClass(
        'long string that will definitely trigger a line break'
    ).method(
        'long argument string that should come on the next line',
        param1=1,
        param2=2.0
    )
    """
    linewidth = np.get_printoptions()['linewidth']

    # Part up to the method name
    if (len(inst_str) + 1 + len(meth_str) + 1 <= linewidth or
            '(' not in inst_str):
        init_parts = [inst_str, meth_str]
        # Length of the line to the end of the method name
        meth_line_start_len = len(inst_str) + 1 + len(meth_str)
    else:
        # TODO(kohr-h): use `maxsplit=1` kwarg, not supported in Py 2
        left, rest = inst_str.split('(', 1)
        right, middle = rest[::-1].split(')', 1)
        middle, right = middle[::-1], right[::-1]
        if middle.startswith('\n') and middle.endswith('\n'):
            # Already on multiple lines
            new_inst_str = inst_str
        else:
            new_inst_str = '(\n'.join([left, indent(middle)]) + '\n)' + right

        # Length of the line to the end of the method name, consisting of
        # ')' + '.' + <method name>
        meth_line_start_len = 1 + 1 + len(meth_str)
        init_parts = [new_inst_str, meth_str]

    # Method call part
    arg_str_oneline = ', '.join(arg_strs)
    if meth_line_start_len + 1 + len(arg_str_oneline) + 1 <= linewidth:
        meth_call_str = '(' + arg_str_oneline + ')'
    elif not arg_str_oneline:
        meth_call_str = '(\n)'
    else:
        if allow_mixed_seps:
            arg_seps = _separators(arg_strs, linewidth - 4)  # indented
        else:
            arg_seps = [',\n'] * (len(arg_strs) - 1)

        full_arg_str = ''
        for arg_str, sep in zip_longest(arg_strs, arg_seps, fillvalue=''):
            full_arg_str += arg_str + sep

        meth_call_str = '(\n' + indent(full_arg_str) + '\n)'

    return '.'.join(init_parts) + meth_call_str
